package de.isse.kiv

import kiv.fileio.file.load_obj
import kiv.project.Devgraph
import org.eclipse.core.resources.IFile
import org.eclipse.core.resources.IProject
import org.eclipse.core.resources.IResource
import org.eclipse.core.resources.IResourceChangeEvent
import org.eclipse.core.resources.IResourceChangeListener
import org.eclipse.core.resources.IResourceDelta
import org.eclipse.core.resources.IWorkspace
import org.eclipse.core.resources.ResourcesPlugin
import org.eclipse.core.resources.refresh.IRefreshMonitor
import org.eclipse.core.resources.refresh.IRefreshResult
import org.eclipse.core.resources.refresh.RefreshProvider
import org.eclipse.core.runtime.CoreException
import org.eclipse.core.runtime.IProgressMonitor
import org.eclipse.core.runtime.IStatus
import org.eclipse.core.runtime.Path
import org.eclipse.core.runtime.Status
import org.eclipse.core.runtime.jobs.Job
import de.isse.kiv.project.Nature
import de.isse.kiv.resources.ProjectModel
import de.isse.kiv.resources.ResourceProperties._
import de.isse.kiv.ui.Console
import scala.collection.JavaConversions._

class DevgraphWatch extends RefreshProvider with IResourceChangeListener {

  private def addWatch(project: IProject, devgraph: IFile) {
    try {
      if (!project.hasNature(Nature.NATURE_ID)) return
    } catch {
      case e: CoreException => return
    }
    if (devgraph == null) return
    if (devgraph.hasMonitor) return

    val monitor = installMonitor(devgraph, null)
    if (monitor == null) return
    devgraph.setMonitor(monitor)

    object job extends Job("touch" + devgraph.getFullPath) {
      override def run(monitor: IProgressMonitor): IStatus = {
        try {
          devgraph.touch(monitor)
        } catch {
          case e: CoreException =>
            Console.error.println("touch on", devgraph.getFullPath, "failed:", e.getMessage)
        }
        return Status.OK_STATUS
      }
    }

    job.schedule()

    Console.debug.println("installing devgraph watch for project", project.getName)
  }

  private def removeWatch(project: IProject, devgraph: IFile) {
    try {
      if (!project.hasNature(Nature.NATURE_ID)) return
    } catch {
      case e: CoreException => return
    }
    if (devgraph == null) return
    val monitor = devgraph.unsetMonitor
    if (monitor == null) return
    monitor.unmonitor(devgraph)
    Console.debug.println("removing devgraph watch for project", project.getName)
  }

  private def fixWatch(delta: IResourceDelta, project: IProject, devgraph: IFile) {
    val kinds = delta.getKind
    val flags = delta.getFlags
    if ((kinds & IResourceDelta.ADDED) != 0) {
      addWatch(project, devgraph)
    } else if ((kinds & IResourceDelta.REMOVED) != 0) {
      removeWatch(project, devgraph)
    } else if ((flags & IResourceDelta.OPEN) != 0) {
      if (project.isOpen) addWatch(project, devgraph) else removeWatch(project, devgraph)
    }
  }

  def reloadDevgraph(project: IProject, path: String) {
    val model = project.projectModel
    try {
      val dvg = load_obj(None, path).asInstanceOf[Devgraph]
      model.devgraph = dvg
      Console.debug.println("reloaded devgraph for project", project.getName)
    } catch {
      case e: Exception =>
        Console.error.println("failed to reload devgraph for project ", project.getName, ": ", e.getMessage)
    }
  }

  def maybeReloadDevgraph(delta: IResourceDelta, project: IProject, file: IFile) {
    if (!file.exists) return
    val unsync = delta.findMember(new Path(DEVGRAPH)) != null
    if (unsync) {
      val path = file.getLocation.toOSString
      reloadDevgraph(project, path)
    }
  }

  override def resourceChanged(event: IResourceChangeEvent) {
    val delta = event.getDelta
    if (delta == null) return
    for (child <- delta.getAffectedChildren) {
      child.getResource match {
        case project: IProject =>
          try {
            if (project.hasNature(Nature.NATURE_ID)) {
              val devgraph = project.devgraphFile
              if (devgraph != null) {
                fixWatch(child, project, devgraph)
                maybeReloadDevgraph(child, project, devgraph)
              }
            }
          } catch {
            case e: CoreException =>
          }
      }
    }
  }

  def start() {
    ResourcesPlugin.getWorkspace.addResourceChangeListener(this)
    val workspace = ResourcesPlugin.getWorkspace
    val projects = workspace.getRoot.getProjects
    for (project <- projects) {
      val devgraph = project.devgraphFile
      addWatch(project, devgraph)
    }
  }

  def stop() {
    val workspace = ResourcesPlugin.getWorkspace
    val projects = workspace.getRoot.getProjects
    for (project <- projects) {
      val devgraph = project.devgraphFile
      removeWatch(project, devgraph)
    }
    workspace.removeResourceChangeListener(this)
  }

  override def installMonitor(resource: IResource, result: IRefreshResult): IRefreshMonitor = {
    createPollingMonitor(resource)
  }
}
