package de.isse.kiv

import java.rmi.AlreadyBoundException
import java.rmi.RemoteException
import java.rmi.registry.LocateRegistry
import java.rmi.server.UnicastRemoteObject

import org.eclipse.core.resources.IFile
import org.eclipse.core.resources.IResource
import org.eclipse.core.resources.ResourcesPlugin
import org.eclipse.core.runtime.CoreException
import org.eclipse.core.runtime.IPath
import org.eclipse.core.runtime.Path
import org.eclipse.jface.wizard.WizardDialog
import org.eclipse.swt.widgets.Display
import org.eclipse.ui.PlatformUI
import org.eclipse.ui.editors.text.IEncodingSupport
import org.eclipse.ui.ide.IDE
import org.eclipse.ui.part.FileEditorInput
import org.eclipse.ui.texteditor.ITextEditor

import de.isse.kiv.resources.ResourceLookup
import de.isse.kiv.resources.ResourceProperties._
import de.isse.kiv.resources.ResourceProperties.toResourceProperties
import de.isse.kiv.ui.Console
import de.isse.kiv.ui.MessageWindow
import de.isse.kiv.ui.editors.FileEditor
import de.isse.kiv.ui.wizards.ImportProjectWizard
import kiv.editor._
import org.eclipse.core.resources.IResourceChangeListener
import org.eclipse.core.resources.IResourceChangeEvent
import java.rmi.NotBoundException
import org.eclipse.core.resources.IProject
import org.eclipse.core.resources.IResourceDelta
import kiv.expr.Op
import de.isse.kiv.project.Definition

@remote
object KIVEditorServer extends UnicastRemoteObject with kiv.editor.EditorServer {
  private var bound = false
  private lazy val registry = LocateRegistry.createRegistry(EditorServer.RMI_REGISTRY_PORT)
  private var kivCallbacks: List[KivCallback] = Nil

  // Register in order to be notified of resource changes, in order to inform KIV
  // Triggered before every auto-build would be triggered (even if auto-build is disabled)
  ResourcesPlugin.getWorkspace.addResourceChangeListener(new IResourceChangeListener {
    override def resourceChanged(event: IResourceChangeEvent) {
      val delta = event.getDelta
      def getSpecifiationsAndSequents(delta: IResourceDelta): List[IResource] = {
        val resource = delta.getResource
        val thisres = if (((delta.getFlags & IResourceDelta.CONTENT) == IResourceDelta.CONTENT) && (resource.isSpecification || resource.isSequents)) {
          resource :: Nil
        } else
          Nil
        thisres ++ delta.getAffectedChildren.flatMap { getSpecifiationsAndSequents }
      }
      val resources = getSpecifiationsAndSequents(delta)
      if (!resources.isEmpty) {
        Console.debug.println("Resources " + resources.map {res => res.getParent.getName + "/" + res.getName }.mkString("(", ", ", ")") + " have changed, informing KIV")

        // Tell KIV about the files
        val absolutePaths = resources.map { _.getLocation.toFile.getAbsolutePath }
        KIVEditorServer.callback {
          _.fileChange(absolutePaths)
        }
      }
    }
  }, IResourceChangeEvent.POST_CHANGE)

  def bind() {
    try {
      assert(!bound)
      registry.bind(EditorServer.RMI_NAME, this)
      bound = true
    } catch {
      case e: RemoteException       =>
      case e: AlreadyBoundException =>
    }
  }

  def unbind() {
    assert(bound)
    registry.unbind(EditorServer.RMI_NAME)
  }

  def toggle() {
    if (bound) unbind() else bind()
  }

  private def refreshPath(path: IPath) {
    val root = ResourcesPlugin.getWorkspace.getRoot
    Console.debug.println("trying to refresh path " + path.toOSString())
    for (project <- root.getProjects) {
      val location = project.projectLocation
      if (location.isPrefixOf(path)) {
        project.refreshLocal(IResource.DEPTH_INFINITE, null)
        Console.debug.println("  refreshing project \"" + project.getName() + "\"")
      }
    }
  }

  def registerCallback(callback: KivCallback) {
    kivCallbacks ::= callback
    Console.debug.println("RMI: KIV client registered")
  }

  def unregisterCallback(callback: KivCallback) {
    kivCallbacks = kivCallbacks.filterNot { _ == callback }
    Console.debug.println("RMI: KIV client unregistered")
  }

  def callback(f: KivCallback => Unit) {
    kivCallbacks.foreach { callback =>
      try {
        f(callback)
      } catch {
        case _: RemoteException | _: NotBoundException =>
      }
    }
  }

  def openTheorem(file: String, name: String) {
    object open extends Runnable {
      override def run() {

        val workspace = ResourceLookup.activeWorkspace
        if (workspace == null) return
        val workspaceRoot = workspace.getRoot

        val path = new Path(file)
        val files = workspaceRoot.findFilesForLocationURI(path.toFile.toURI)
        if (files.isEmpty) return

        val _file = files(0)
        val fileToOpen = _file.asInstanceOf[IFile]

        if (fileToOpen == null || !fileToOpen.exists) return ;

        val desc = PlatformUI.getWorkbench.getEditorRegistry.getDefaultEditor(fileToOpen.getName)
        val page = ResourceLookup.activePage

        val editor = page.openEditor(new FileEditorInput(fileToOpen), desc.getId).asInstanceOf[(ITextEditor)]

        // TODO: use the parsed List[Theorem] and the location there
        val fileModel = fileToOpen.fileModel
        if (fileModel != null) {

          //Try and find token, followed by a colon
          val tokens = fileModel.findTokens(name)
          //test all occurrences in the document for a trailing colon

          val editorContent = editor.getDocumentProvider.getDocument(editor.getEditorInput).get

          val symbol = tokens.find(t => t.location.position + t.location.length + 1 < editorContent.length()
            && editorContent.substring(t.location.position, t.location.position + t.location.length + 1) == (name + ":"))

          if (symbol.nonEmpty)
            editor.selectAndReveal(symbol.get.location.position, symbol.get.location.length)

        }

      }
    }
    PlatformUI.getWorkbench.getDisplay.asyncExec(open);

  }

  /** Get IProject from absolute path */
  def getProject(filename: String): Option[IProject] = {
    val pathToOpen = new Path(filename)
    val fileToOpen = pathToOpen.toFile()
    if (fileToOpen.exists && fileToOpen.isFile) {
      val page = ResourceLookup.activePage
      refreshPath(pathToOpen)
      ResourceLookup.getProject(fileToOpen.getAbsolutePath)
    } else {
      Console.error.println("Error: file \"" + filename + "\" does not exist.\n")
      MessageWindow.showMessage_OK("File Error", "File \"" + filename + "\" does not exist!")
      None
    }
  }

  def openEditor(filename: String) {
    object open extends Runnable {
      override def run() {
        getProject(filename) match {
          case Some(project) =>
            project.open(null)
            project.refreshLocal(IResource.DEPTH_INFINITE, null)
          case None =>
            val answer = MessageWindow.showMessage_YES_NO("Import Project?", "This file can't be found in any open project. Do you want to import the corresponding project?")

            if (answer == 1) {
              //Second Button (Yes) clicked -> Open Import Dialog
              val descriptor = PlatformUI.getWorkbench.getImportWizardRegistry.findWizard(KIVPlugin.IMPORT_WIZARD_ID)

              if (descriptor != null) {
                val wizard = descriptor.createWizard

                val wd = new WizardDialog(Display.getCurrent.getActiveShell, wizard)
                wd.setTitle(wizard.getWindowTitle)

                val importProjectWizard = wizard.asInstanceOf[ImportProjectWizard]

                //TODO: Finden vom Project root noch umbauen???
                var path: IPath = new Path(filename)
                if (filename.contains("specs")) {

                  val x = path.lastSegment()
                  while (path.lastSegment() != "specs")
                    path = path.removeLastSegments(1)
                  path = path.removeLastSegments(1)
                }
                importProjectWizard.setPath(path.toString())

                wd.open
              }
            }
        }

        val file = new Path(filename).toFile
        Console.debug.println("opening file " + file.toURI())
        val part = IDE.openEditor(ResourceLookup.activePage, file.toURI(), FileEditor.FILE_EDITOR_ID, true)
        val encodingSupport = part.getAdapter(classOf[IEncodingSupport]).asInstanceOf[IEncodingSupport]
        encodingSupport.setEncoding("UTF-8")
      }
    }

    PlatformUI.getWorkbench.getDisplay.asyncExec(open)
  }

  def updateDevgraph(filename: String) {
    Console.error.println("devgraph update for " + filename + " has been notified")
    object update extends Runnable {
      override def run() {
        try {
          ResourceLookup.getProject(filename) match {
            case Some(project) =>
              KIVPlugin.getDefault.watch.reloadDevgraph(project, filename)
            case None =>
              Console.error.println("could not find project for path \"", filename, "\"")
          }
        } catch {
          case e: CoreException =>
            Console.error.println("devgraph update failed: " + e)
        }
      }
    }

    PlatformUI.getWorkbench.getDisplay.asyncExec(update);
  }

  def gotoDefinition(filename: String, op: Op) {
    Console.debug.println("going to definition: " + op + " in file " + filename)
    object gotodef extends Runnable {
      override def run() {
        ResourceLookup.getFile(filename).map { file =>
          Console.debug.println("found file " + file + " " + file.exists())
          Definition.gotoDefinitionAction(file, op).foreach { action =>
            Console.debug.println("found action")
            action()
          }
        }
      }
    }

    PlatformUI.getWorkbench.getDisplay.asyncExec(gotodef)
  }
}
