package de.isse.kiv

import java.io.BufferedReader
import java.io.File
import java.io.InputStream
import java.io.InputStreamReader
import org.eclipse.core.resources.IProject
import org.eclipse.core.runtime.IPath
import org.eclipse.core.runtime.Path
import org.eclipse.core.runtime.Platform
import de.isse.kiv.resources.ResourceLookup
import de.isse.kiv.resources.ResourceProperties._
import de.isse.kiv.ui.Console
import de.isse.kiv.ui.MessageWindow
import KIVSystem._

object KIVSystem {
  val SCALA_V20_PLUGIN_ID = "org.scala-lang.scala-library"
  val SCALA_V20_PARSER_PLUGIN_ID = "org.scala-lang.modules.scala-parser-combinators"
  val KIV_PLUGIN_ID = "kiv"
  val KODKOD_PLUGIN_ID = "kodkod.plugin"
  val JGRAPHX_PLUGIN_ID = "jgraphx.plugin"
  val LIBRARY_PLUGIN_ID = "kiv.lib.basic"

  var running = false
  var process: Process = _
  var thread: Thread = _

  def launch(args: String*) {
    val options = if (Settings.isExpertMode) "-expert" :: Nil else Nil
    if (Settings.isForkRequired)
      launchProcess(options ++ args: _*)
    else
      launchThread(options ++ args: _*)
  }

  def workon(project: IProject) {
    val location = project.projectLocation
    val osPath = location.toOSString()
    if (!location.toFile().exists())
      MessageWindow.showMessage_OK("Directory Error", "Directory \"" + osPath + "\" does not exist!")
    else
      launch(osPath)
  }

  def startkiv() {
    launch()
  }

  class StreamRedirection(private var in: InputStream, private var out: Console.Output) extends Thread {
    override def run() {
      running = true
      try {
        val reader = new BufferedReader(new InputStreamReader(in))
        var line: String = null
        while (running) {
          line = reader.readLine()
          if (line == null)
            running = false
          else
            out.synchronized {
              out.println(line)
              out.flush()
            }
        }
      } catch {
        case e: Exception => e.printStackTrace()
      }
      process = null
      running = false
    }
  }

  private def kivMain(args: Array[String]) {
    val expert = false;
    jkiv.GlobalProperties.initStatic(expert)

    // Run the GUI in a new thread
    new Thread(new jkiv.KIVSystem(Thread.currentThread())).start()

    // Run the KIV core in this thread
    try {
      if (args.isEmpty)
        kiv.kivstate.cosi.start()
      else
         kiv.kivstate.cosi.kiv_select_project(args(0))
    } catch {
      case kiv.basic.Exiterror =>
    }
  }

  private def launchThread(args: String*) {
    thread = new Thread() {
      override def run() {
        running = true
        Console.out.println("Starting KIV thread...")
        thread = this
        try {
          kivMain(args.toArray)
        } catch {
          case e: Exception => e.printStackTrace()
        }
        thread = null
        running = false
        Console.out.println("KIV thread terminated.")
      }
    }
    thread.start()
  }
  
  def pluginJar(plugin: String, file: String) = {
      val path = ResourceLookup.pluginPath(plugin)
      val jar = path.append(file)
      jar.toOSString
  }

  private def optionalPluginFile(plugin: String, file: String) = {
    val path = ResourceLookup.pluginPath(plugin)
    if (path == null)
      None
    else
      Some(path.append(file).toOSString())
  }
  
  def scalaLibraryJar = {
      // NOTE: works with 2.11, but might not with previous versions
      val scalaBundle = Platform.getBundle(SCALA_V20_PLUGIN_ID)
      org.eclipse.core.runtime.FileLocator.getBundleFile(scalaBundle).getPath()
  }

  def scalaParserLibraryJar = {
    val scalaBundle = Platform.getBundle(SCALA_V20_PARSER_PLUGIN_ID)
    org.eclipse.core.runtime.FileLocator.getBundleFile(scalaBundle).getPath()
  }

  private def launchProcess(args: String*) {
      val jars = List(
          scalaLibraryJar,
          scalaParserLibraryJar,
          pluginJar(KODKOD_PLUGIN_ID, "kodkod.jar"),
          pluginJar(KODKOD_PLUGIN_ID, "org.sat4j.core.jar"),
          pluginJar(JGRAPHX_PLUGIN_ID, "jgraphx.jar"),
          pluginJar(KIV_PLUGIN_ID, "kiv.jar")
      )

      val kivHome = Settings.kivHome
      val classpath = jars.mkString(File.pathSeparator)

      val javaCommand = Array(
        java.toOSString(),
        "-Xss" + Settings.stacksize,
        "-classpath",
        classpath,
        "Main")

      val kivCommand = javaCommand ++ args
      
      val escapedKivCommand = kivCommand.map("\"" + _ + "\"")
      Console.debug.println("Running: " + escapedKivCommand.mkString(" ") + ", KIVHOME=\"" + kivHome + "\"")
      
      val builder = new ProcessBuilder(kivCommand : _*)
      val env = builder.environment()

      if (Settings.useInternalEditor) env.remove("KIVEDITOR")
      env.put("KIVHOME", kivHome)

      builder.redirectErrorStream()
      process = builder.start()
      new StreamRedirection(process.getInputStream, Console.debug).start()
      new StreamRedirection(process.getErrorStream, Console.debug).start()
  }

  def java = new Path(System.getProperty("java.home")).append("bin").append("java")
}
