package de.isse.kiv.resources

import kiv.project.Devgraph
import kiv.spec._
import kiv.prog._
import kiv.expr._
import de.isse.kiv.source.parser.KIVParser
import org.eclipse.core.resources.IFile
import kiv.signature.Sigentry
import scala.annotation.tailrec
import kiv.signature._
import kiv.util.ScalaExtensions._
import de.isse.kiv.ui.Console
import de.isse.kiv.resources.ResourceProperties._
import kiv.util.MultiGraph
import kiv.util.MultiGraph._
import kiv.util.DefaultEdge
import kiv.parser.PreOp
import kiv.parser.PreTyCo
import kiv.module.Isexpr

class ProjectModel extends de.isse.kiv.Observable[ProjectModel] {
  private var _needsFullBuild: Boolean = _

  private var _dvg: Option[Devgraph] = None
  private var _devgraph: Map[String, Spec] = Map()
  private var _current: Map[String, Spec] = Map()
  private var _specList: List[(String, Spec)] = Nil
  private var _subSpecs: Map[String, List[String]] = Map()
  private var _subSpecsReversed: Map[String, List[String]] = Map()

  def specList = _specList
  def devgraph = _devgraph
  def subSpecs = _subSpecs
  def subSpecsReversed = _subSpecsReversed
  def specNames = specList.map(_._1)

  def devgraph_=(dvg: Devgraph) {
    _dvg = Some(dvg)
    val newDevgraph = dvg.getokdevspecs.toMap

    val devgraphMultigraph = MultiGraph.fromDevgraph(dvg)
    val reversedMultigraph = new MultiGraph[DevgraphNode, DefaultEdge[DevgraphNode]]()
    devgraphMultigraph.foreachNode { node => reversedMultigraph.add(node) }
    devgraphMultigraph.foreachEdge { edge => reversedMultigraph.add(new DefaultEdge[DevgraphNode](edge.target, edge.source)) }

    // Only keep versions of specs that
    // are potentially newer in Eclipse than in KIV,
    // i.e, that haven't changed in KIV so far.

    // Note: don't use filterKeys
    // it's lazy and will stack up with each update
    // (the filter was evaluated during typing)
    _current = _current.filter {
      case (name, _) =>
        _devgraph.get(name) == newDevgraph.get(name)
    }

    _devgraph = newDevgraph
    _needsFullBuild = true
    _specList = (_devgraph ++ _current).toList

    _subSpecs = dvg.devunits.map {
      du =>
        (du.unitname.name, du.unit_get_used_units)
    }.toMap

    _subSpecsReversed = reversedMultigraph.getNodes.map {
      dn =>
        (dn.unit.unitname.name, getTargetedDevgraphNodes(dn, reversedMultigraph))
    }.toMap

    notify(this)
  }

  private def getTargetedDevgraphNodes(node: DevgraphNode, graph: MultiGraph[DevgraphNode, DefaultEdge[DevgraphNode]]): List[String] = {
    graph.getEdges.filter { edge => edge.source == node }.map { edge => edge.target.unit.unitname.name }.toList
  }

  def updateSpec(name: String, spec: Spec) = {
    _current += name -> spec
    _specList = (_devgraph ++ _current).toList
    notify(this)
  }

  def getSpec(name: String): Option[Spec] = _current.get(name) match {
    case Some(spec) => Some(spec)
    case None       => _devgraph.get(name)
  }

  /**
   * Return the specification which defines this sort
   * Note that since operation may be renamed the original sort is also returned
   */
  def getDefinition(specname: String, tyco: TyCo): Option[(TyCo, String)] = {
    def rec(tyco: TyCo): Option[(TyCo, String)] = {
      _subSpecs.getOrElse(specname, Nil).collectFirst { subspec =>
        getDefinition(subspec, tyco) match {
          case Some(res) =>
            res
        }
      }
    }

    getSpec(specname).map { spec =>
      spec match {
        case _: Actualizedspec | _: Renamedspec =>
          val origtyco = spec.morphism.symrenlist.filterType[Sortren].find { _.rensort == tyco.toType }.map { _.sort }.getOrElse(tyco)
          rec(tyco).orElse(rec(origtyco))
        case _: Instantiatedspec =>
          val origtyco = (spec.mapping.symmaplist ++ spec.mapping.extsymmaplist).filterType[Sortmap].find { _.maptypelist == tyco.toType :: Nil }.map { _.sort }.getOrElse {
            spec.mapping.symrenlist.filterType[Sortren].find { _.rensort == tyco.toType }.map { _.sort }.getOrElse(tyco)
          }
          return rec(origtyco)
        case _ =>
          val ownsig = spec.top_signature
          if (ownsig.sortlist.contains(tyco))
            Some((tyco, specname))
          else
            rec(tyco)
      }
    }.flatten
  }

  /**
   * Return the specification which defines this operation
   * Note that since operation may be renamed the original operation is also returned
   */
  def getDefinition(specname: String, op: Op): Option[(Op, String)] = {
    def rec(op: Op): Option[(Op, String)] = {
      _subSpecs.getOrElse(specname, Nil).collectFirst { subspec =>
        getDefinition(subspec, op) match {
          case Some(res) =>
            res
        }
      }
    }

    getSpec(specname).map { spec =>
      spec match {
        case _: Actualizedspec | _: Renamedspec =>
          val origop = spec.morphism.symrenlist.filterType[Opren].find { _.renop == op }.map { _.op }.getOrElse(op)
          rec(origop)
        case _: Instantiatedspec =>
          val origop = (spec.mapping.symmaplist ++ spec.mapping.extsymmaplist).filterType[Opmap].find { _.mapexprorproclist == Isexpr(op.toInstOp) :: Nil }.map { _.op }.getOrElse {
            spec.mapping.symrenlist.filterType[Opren].find { _.renop == op }.map { _.op }.getOrElse(op)
          }
          return rec(origop)
        case _ =>
          val ownsig = spec.top_signature
          if ((ownsig.oplist ++ ownsig.poplist).contains(op))
            Some((op, specname))
          else
            rec(op)
      }
    }.flatten
  }

  /**
   *  Find the declaration for the procedure
   *  Note that since procedures may be renamed the returned declaration could define a procedure
   *  with a different symbol
   */
  def getDeclaration(specname: String, proc: AnyProc): Option[(Anydeclaration, String)] = {
    def rec(proc: AnyProc): Option[(Anydeclaration, String)] = {
      _subSpecs.getOrElse(specname, Nil).collectFirst { subspec =>
        getDeclaration(subspec, proc) match {
          case Some(res) =>
            res
        }
      }
    }

    getSpec(specname) match {
      case None =>
        None
      case Some(spec) =>
        val decls = spec match {
          case _: Basicspec | _: Enrichedspec | _: Genspec | _: ASMspec | _: DataASMSpec | _: Complexspec =>
            spec.decllist
          case _: Renamedspec | _: Actualizedspec =>
            val origproc = spec.morphism.symrenlist.filterType[Procren].find { _.renproc == proc }.map { _.proc }.getOrElse(proc)
            return rec(origproc)
          case spec: Instantiatedspec =>
            val origproc = (spec.mapping.symmaplist ++ spec.mapping.extsymmaplist).filterType[Procmap].find { _.mapproc == proc }.map { _.proc }.getOrElse {
              spec.mapping.symrenlist.filterType[Procren].find { _.renproc == proc }.map { _.proc }.getOrElse { proc }
            }
            return rec(origproc)
          case spec: ReducedDataASMSpec =>
            spec.reduceddataasm.decllist.find { _.declprocdecl.proc == proc } match {
              case Some(decl) =>
                spec.reductionspec.dataasm.decllist.find { _.declname == decl.declname } match {
                  case Some(origdecl) =>
                    return rec(origdecl.declprocdecl.proc)
                  case None =>
                    return rec(proc)
                }
              case None =>
                return rec(proc)
            }
          case _ =>
            Nil
        }
        decls.find { _.declprocdecl.proc == proc } match {
          case Some(procdecl) =>
            Some(procdecl, specname)
          case None =>
            rec(proc)
        }
    }
  }

  def needsFullBuild: Boolean = {
    val res = _needsFullBuild
    _needsFullBuild = false
    res
  }

  def getProcMap(file: IFile): Map[String, NewProc] = {
    val spec = getSpec(file)
    val signature = signatureAfter(file)
    val procList0 = signature.proclist.filterType[NewProc].map { proc => (proc.procsym.name, proc) }
    val procList = procList0 ++ (spec match {
      case Some(asm: DataASMSpec) =>
        val prefix = asm.name.name + "_"
        val len = prefix.length
        signature.proclist.filterType[NewProc].filter { _.procsym.name.startsWith(prefix) }.map { proc => (proc.procsym.name.substring(len), proc) }

      case spec @ Some(_: Actualizedspec | _: Renamedspec) =>
        getOriginalProcsFromRenamings(spec.get.morphism.symrenlist) map { p => (p.procsym.name, p) }

      case Some(inst: Instantiatedspec) =>
        getOriginalProcsFromRenamings(spec.get.mapping.symrenlist) map { p => (p.procsym.name, p) }

      case _ => Nil
    })
    procList.toMap
  }

  private def getOriginalProcsFromRenamings(rens: List[Symren]): List[NewProc] = {
    val procrens = rens.filterType[Procren]
    val origprocs = procrens map (_.proc)
    origprocs.filterType[NewProc]
  }

  def findProc(file: IFile, name: String): Option[NewProc] = getProcMap(file).get(name)

  /** Signature before parsing the file */
  def signatureBefore(file: IFile): Currentsig = signature(if (file.isSpecification) None else getSpec(file))

  /** Signature of the parsed file */
  def signatureAfter(file: IFile): Currentsig = signature(getSpec(file))

  private def signature(spec: Option[Spec]): Currentsig = {
    val sig = spec match {
      case None       => Signature.empty_signature
      case Some(spec) => spec.specsignature
    }
    sig.toCurrentsigWithKeep
  }

  /** Get the latest specification of the file */
  def getSpec(file: IFile): Option[Spec] = getSpec(file.specName)

  def getReducedDeclaration(file: IFile, proc: NewProc): Option[(Anydeclaration, Fpl)] = {
    val (decl, subspec) = getDeclaration(file.specName, proc) match {
      case Some(res) => res
      case None      => return None
    }
    val fpl = decl.declprocdecl.fpl
    val spec = getSpec(file)
    spec match {
      case Some(asmspec: DataASMSpec) if file.isSpecification && (subspec == file.specName || (_subSpecs.contains(file.specName) && _subSpecs(file.specName).contains(subspec) && getSpec(subspec).get.isInstanceOf[DataASMSpec])) =>
        // call into this machine or submachine
        val state = asmspec.fullstate
        Some(decl, Fpl(fpl.fvalueparams.diff(state), fpl.fvarparams.diff(state), fpl.foutparams.diff(state)))
      case _ =>
        Some(decl, fpl)
    }
  }

  def getProcCompletions(file: IFile, name: String): List[NewProc] = {
    val asmname = getASMName(file)
    signatureAfter(file).proclist.filterType[NewProc].filter { proc => proc.procsym.name.startsWith(name) || (asmname != None && proc.procsym.name.startsWith(asmname.get.name + "_" + name)) }
  }

  def getASMName(file: IFile): Option[Symbol] = {
    if (!file.isSpecification)
      None
    else getSpec(file) match {
      case Some(dataasm: DataASMSpec) =>
        Some(dataasm.name)
      case _ =>
        None
    }
  }

  def getSpecLibrary(specname: String): Option[String] = {
    if (_dvg == None) return None
    _dvg.get.devspeclist.find { devspec => devspec.libp && devspec.specname == specname }.map { _.libname }
  }
}
