package de.isse.kiv.resources

import kiv.parser.Terminal
import kiv.parser.PreSpecification

import kiv.parser.Terminals.T_AXIOMS
import kiv.spec.Spec
import kiv.spec.Theorem
import de.isse.kiv.ui.Console
import kiv.signature.Currentsig
import kiv.prog.AnyProc
import kiv.util.ScalaExtensions._
import de.isse.kiv.source.parser.KIVToken
import de.isse.kiv.source.parser.RawToken
import de.isse.kiv.source.analyzers._
import kiv.parser.PreFl1
import kiv.parser.PreFl3
import kiv.parser.PreExpr
import de.isse.kiv.source._
import kiv.parser.PreOp
import org.eclipse.core.resources.IFile
import org.eclipse.ui.texteditor.IDocumentProvider
import org.eclipse.ui.editors.text.TextFileDocumentProvider
import kiv.expr.Op
import de.isse.kiv.source.parser.SpecMapsAnalyzer
import de.isse.kiv.source.parser.OperationToken
import de.isse.kiv.source.parser.KIVTokenizer
import kiv.parser.PreTyCo
import de.isse.kiv.source.parser.SortDefinitionToken
import kiv.expr.TyCo
import kiv.expr.FormulaPattern._

object FileModel {
  class RegionString(val text: String, var start: Int, var end: Int) {
    override def toString(): String = text

    def setRange(start: Int, end: Int) = {
      this.start = start
      this.end = end
    }
  }
}

class FileModel extends de.isse.kiv.Observable[FileModel] {

  private var spec: Spec = _
  private var prespec: PreSpecification = _
  private var _theorems: List[Theorem] = Nil
  private var specAnalyzer: SpecMapsAnalyzer = _

  def theorems = _theorems
  def tokens = _tokens

  /** The token list is only valid for files already opened in the editor */
  private var _tokens: List[KIVToken] = Nil

  def updateTokens(tokens: List[KIVToken]) {
    val donotify = (_tokens == tokens)
    _tokens = tokens
    _axiomToken = findToken(T_AXIOMS)
    if (donotify)
      notify(this)
  }

  private var _axiomToken: Option[KIVToken] = None

  def axiomToken = _axiomToken

  def isPositionBeforeAxiomToken(position: Int): Boolean = _axiomToken match {
    case None                     => true
    case Some(kivToken: KIVToken) => position < (kivToken.location.position until (kivToken.location.position + kivToken.location.length)).start
  }

  def tokenAt(offset: Int) = {
    tokens.find { x => (x.location.position until (x.location.position + x.location.length)) contains offset }
  }

  def findToken(id: Short) = {
    tokens.filterType[RawToken].find { _.terminal.getId == id }
  }

  def findFirstOccurrence(op: Op): Option[Range] = {
    _tokens.filterType[OperationToken].find { _.op.op == op }.map { _.location.toRange }
  }

  def findDefinitionToken(s: String): Option[KIVToken] = {
    val defToken = _tokens.find(t => t.stringPresentation == s)
    defToken match {
      case Some(kivtoken: KIVToken) => Some(kivtoken)
      case None                     => None
    }
  }

  def findDefinitionRange(tyco: TyCo): Option[Range] = {
    _tokens.filterType[SortDefinitionToken].find { _.tycodef.symloc.sym == tyco.tycosym }.map { _.location.toRange }
  }

  def findDefinitionRange(proc: AnyProc): Option[Range] = {
    spec.decllist.find { decl => decl.declprocdecl.proc == proc && decl.location != null }.map { _.location.toRange }
  }

  def findDefinitionRange(op: Op, file: IFile): Option[Range] = {
    if (tokens.isEmpty)
      tokenizeTokensForFile(file)
    if (prespec.theorems == null)
      return None

    // try to find an equation/equivalence that defines the operation, if none exists choose first occurrence
    prespec.theorems.find { theorem =>
      theorem.theoremseq.suc match {
        case definition :: Nil =>
          val lhsrhs = definition match {
            case Equiv(lhs, rhs) =>
              Some(lhs, rhs)
            case Equal(lhs, rhs) =>
              Some(lhs, rhs)
            case Imp(_, Equiv(lhs, rhs)) =>
              Some(lhs, rhs)
            case Imp(_, Equal(lhs, rhs)) =>
              Some(lhs, rhs)
            case _ =>
              None
          }
          lhsrhs match {
            case Some((lhs, rhs)) if lhs.ops_of_expr.contains(op) =>
              true
            case _ =>
              false
          }
        case _ =>
          false
      }
    }.orElse(prespec.theorems.find { theorem =>
      val suc = theorem.theoremseq.suc
      suc.flatMap { _.ops_of_expr }.contains(op)
    }).map { _.location.toRange }
  }

  def tokenizeTokensForFile(file: IFile) {
    val provider: IDocumentProvider = new TextFileDocumentProvider();
    provider.connect(file);
    val document = provider.getDocument(file)
    KIVTokenizer.parse(document.get, file, prespec)
  }

  private def axiomTokenBeforeKivToken(kivToken: KIVToken): Boolean = {
    val axiomToken = tokens.find { token => token.stringPresentation == "axioms" }
    axiomToken match {
      case Some(token) => token.location.position < kivToken.location.position
      case _           => false
    }
  }

  private def beforebeforeToken(list: List[KIVToken], token: KIVToken): KIVToken = {
    val index = list.indexOf(token)
    list(index - 2)
  }

  private def beforeToken(list: List[KIVToken], token: KIVToken): KIVToken = {
    val index = list.indexOf(token)
    list(index - 1)
  }

  def findTokens(s: String) = {
    _tokens.filter(t => t.stringPresentation == s)
  }

  def update(parsed: Any) {
    parsed match {
      case spec: Spec =>
        this.spec = spec
      case theorems: List[Theorem @unchecked] if theorems.forall(_.isInstanceOf[Theorem]) =>
        this._theorems = theorems
      case prespec: PreSpecification =>
        this.prespec = prespec
      case _ =>
        return
    }
    notify(this)
  }

  def updateSpecAnalyzer(specAnalyzer: SpecMapsAnalyzer) = {
    this.specAnalyzer = specAnalyzer
  }

  def getSpecAnalyzer: Option[SpecMapsAnalyzer] = {
    if (specAnalyzer != null) Some(specAnalyzer)
    else None
  }

  def getPreSpecificationOrSequents: Option[Either[PreSpecification, List[Theorem]]] = {
    if (prespec != null) Some(Left(prespec))
    else if (theorems != Nil) Some(Right(theorems))
    else None
  }

  def getSpec: Option[Spec] = {
    if (spec != null) Some(spec)
    else None
  }
}
