package de.isse.kiv.resources

import java.net.URL
import org.eclipse.core.resources.IFile
import org.eclipse.core.resources.IProject
import org.eclipse.core.runtime.FileLocator
import org.eclipse.core.runtime.Path
import org.eclipse.jface.resource.ImageDescriptor
import org.eclipse.jface.text.IDocument
import org.eclipse.swt.graphics.Image
import org.eclipse.ui.IEditorPart
import org.eclipse.ui.IEditorReference
import org.eclipse.ui.IFileEditorInput
import org.eclipse.ui.IWorkbenchPage
import org.eclipse.ui.IWorkbenchWindow
import org.eclipse.ui.PlatformUI
import org.eclipse.ui.part.FileEditorInput
import org.eclipse.ui.texteditor.ITextEditor
import de.isse.kiv.KIVPlugin
import de.isse.kiv.resources.ResourceProperties._
import org.eclipse.ui.IEditorInput
import org.eclipse.ui.ide.FileStoreEditorInput
import org.eclipse.core.resources.IFolder
import org.eclipse.jface.text.source.ISourceViewer
import org.eclipse.jface.text.ITextOperationTarget
import org.eclipse.core.runtime.IPath
import org.eclipse.core.resources.ResourcesPlugin
import org.eclipse.core.resources.IResource
import org.eclipse.core.resources.IContainer
import org.eclipse.core.resources.IWorkspace
import org.eclipse.core.filesystem.EFS
import org.eclipse.core.filesystem.IFileStore
import de.isse.kiv.Settings
import java.nio.file.Paths

object ResourceLookup {
  def activeWindow: IWorkbenchWindow = {
    val workbench = PlatformUI.getWorkbench
    if (workbench == null) return null
    workbench.getActiveWorkbenchWindow
  }

  def activeWorkspace: IWorkspace = {
    ResourcesPlugin.getWorkspace
  }

  def activePage: IWorkbenchPage = {
    val window = activeWindow
    if (window == null) return null
    window.getActivePage
  }

  def activeEditor: IEditorPart = {
    val page = activePage
    if (page == null) return null
    page.getActiveEditor
  }

  def activeTextEditor: ITextEditor = {
    activeEditor match {
      case part: ITextEditor =>
        part
      case _ =>
        null
    }
  }

  def activeSourceViewer: ISourceViewer = {
    activeEditor.getAdapter(classOf[ITextOperationTarget]).asInstanceOf[ISourceViewer]
  }

  def document: IDocument = {
    val editor = activeTextEditor
    if (editor == null) return null
    val dp = editor.getDocumentProvider
    if (dp == null) return null
    dp.getDocument(editor.getEditorInput)
  }

  def document(editorref: IEditorReference): IDocument = {
    if (editorref == null) return null

    editorref.getEditor(false) match {
      case editor: ITextEditor =>
        val dp = editor.getDocumentProvider
        if (dp == null) return null
        dp.getDocument(editor.getEditorInput)
      case _ =>
        null
    }
  }

  def documentResource: IFile = {
    val editor = activeEditor
    if (editor == null) return null
    editor.getEditorInput match {
      case input: IFileEditorInput =>
        input.getFile
      case _ =>
        null
    }
  }

  def activeProject: IProject = {
    val file = documentResource
    if (file == null) return null
    file.getProject
  }

  /**
   * Returns the corresponding IFile to an IDocument if it can be found in one of the opened editors
   * @param doc The IDocument
   */
  def iDocumentToIFile(doc: IDocument): Option[IFile] = {

    val editors = allOpenEditors

    editors.foreach { x =>
      if (document(x) == doc) {
        val input = x.getEditorInput
        if (input == null) None
        input match {
          case input: IFileEditorInput =>
            return Some(input.getFile)
          case _ =>
            return None
        }
      }
    }
    None
  }

  def getIFileFromITextEditor(itextEditor: ITextEditor): IFile = {
    if (itextEditor == null) return null
    itextEditor.getEditorInput() match {
      case input: FileEditorInput =>
        input.getFile
      case _ =>
        null
    }
  }

  def allOpenEditors: Array[IEditorReference] = {
    PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getEditorReferences()
  }

  def stockImage(name: String): Image = {
    val workbench = PlatformUI.getWorkbench
    if (workbench == null) return null
    val images = workbench.getSharedImages
    if (images == null) return null
    images.getImage(name)
  }

  lazy val KIVIcon = icon("icons/kiv.gif")
  lazy val KIVIconGreen = icon("icons/kivgreen.gif")
  lazy val KIVError = icon("icons/error.gif")

  def icon(name: String) = {
    val desc = ImageDescriptor.createFromURL(new URL(KIVPlugin.PLUGIN_URL + "/" + name))
    desc.createImage();
  }

  def editedFile(editor: ITextEditor): IFile = {
    editor.getEditorInput match {
      case input: FileEditorInput =>
        input.getFile
      case _ =>
        null
    }
  }

  def specFolder(name: String, project: IProject = activeProject): IFolder = {
    project.getFolder("specs").getFolder(name)
  }

  def specFile(name: String, project: IProject = activeProject): IFile = {
    specFolder(name, project).getFile("specification.utf8")
  }

  implicit class EitherIFileIFileStore(either: Either[IFile, IFileStore]) {
    def fileExists: Boolean = either match {
      case Left(file) =>
        ! file.getProject.isOpen || file.exists()
      case Right(file) =>
        file.fetchInfo().exists()
    }
    def fileModel: FileModel = either match {
      case Left(file) =>
        file.fileModel
      case Right(_) =>
        null
    }
  }

  def specFileOrLibrary(name: String, project: IProject = activeProject): Either[IFile, IFileStore] = {
    val file = specFolder(name, project).getFile("specification.utf8")
    val projectModel = project.projectModel
    projectModel.getSpecLibrary(name) match {
      case None =>
        Left(file)
      case Some(library) =>
        kiv.fileio.ProjectsFile.get.find { _.name == library } match {
          case None =>
            Left(file)
          case Some(projectsEntry) =>
            val kivpath = projectsEntry.path
            val projectpath = kiv.fileio.file.resolveKIVPath(kivpath)
            getProject(projectpath.getAbsolutePath) match {
              case Some(project) =>
                val file = project.getFile(s"specs/${name}/specification.utf8")
                if (file != null)
                  return Left(file)
              case None =>
            }
            val fileStore = EFS.getLocalFileSystem().getStore(projectpath.toURI())
            Right(fileStore.getChild("specs").getChild(name).getChild("specification.utf8"))
        }
    }
  }
  
  def seqFileOrLibrary(name: String, project: IProject = activeProject): Either[IFile, IFileStore] = {
    val file = specFolder(name, project).getFile("sequents.utf8")
    val projectModel = project.projectModel
    projectModel.getSpecLibrary(name) match {
      case None =>
        Left(file)
      case Some(library) =>
        kiv.fileio.ProjectsFile.get.find { _.name == library } match {
          case None =>
            Left(file)
          case Some(projectsEntry) =>
            val kivpath = projectsEntry.path
            val projectpath = kiv.fileio.file.resolveKIVPath(kivpath)
            getProject(projectpath.getAbsolutePath) match {
              case Some(project) =>
                val file = project.getFile(s"specs/${name}/sequents.utf8")
                if (file != null)
                  return Left(file)
              case None =>
            }
            val fileStore = EFS.getLocalFileSystem().getStore(projectpath.toURI())
            Right(fileStore.getChild("specs").getChild(name).getChild("sequents.utf8"))
        }
    }
  }
  
  
  

  def seqFile(name: String, project: IProject = activeProject): IFile = {
    specFolder(name, project).getFile("sequents.utf8")
  }

  def pluginPath(id: String) = {
    val purl = pluginUrl(id)
    if (purl == null)
      null
    else {
      val url = FileLocator.resolve(purl)
      if (url == null)
        null
      else
        new Path(url.getPath)
    }
  }

  def pluginUrl(id: String) = {
    FileLocator.find(new URL("platform:/plugin/" + id))
  }

  def getProject(path: String): Option[IProject] = {
    getProjectName(path).flatMap { name =>
      val workspaceRoot = ResourcesPlugin.getWorkspace.getRoot
      val project = workspaceRoot.getProject(name) // NOTE: only returns a resource handle without checking existence
      if (project != null && project.exists()) Some(project) else None
    }
  }

  def getFile(path: String): Option[IFile] = {
    getProject(path).flatMap { project =>
      project.open(null)
      project.refreshLocal(IResource.DEPTH_INFINITE, null)
      getProjectRelativePath(path).map { project.getFile }
    }
  }

  def getProjectName(path: String): Option[String] = {
    splitProjectPath(path).map {
      case (project, _) =>
        project.lastSegment()
    }
  }

  def getProjectRelativePath(absolutepath: String) = splitProjectPath(absolutepath).map { _._2 }

  /** Splits an absolute path into the part that refers to the project and the remainder */
  def splitProjectPath(absolutepath: String, rest: IPath = null): Option[(IPath, IPath)] = {
    def restPath(name: Option[String]): IPath = name match {
      case None =>
        rest
      case Some(name) =>
        new Path(name).append(rest)
    }

    val path = new Path(absolutepath)
    val file = path.toFile()

    if (file.isDirectory()) {
      val fileNameList = file.listFiles().map(x => x.getName)
      if (fileNameList.contains("devgraph"))
        return Some(new Path(file.getAbsolutePath), restPath(None))
      else {
        if (file.getParent != null) {
          return splitProjectPath(file.getParent, restPath(Some(file.getName)))
        } else
          return null
      }
    } else {
      if (file.getParent != null) {
        return splitProjectPath(file.getParent, restPath(Some(file.getName)))
      } else
        return None
    }
  }
}
