package de.isse.kiv.source

import scala.collection.mutable.ArrayBuffer
import org.eclipse.jface.text.IRegion
import org.eclipse.jface.text.ITextViewer
import org.eclipse.jface.text.Region
import org.eclipse.jface.text.hyperlink.IHyperlink
import org.eclipse.jface.text.hyperlink.IHyperlinkDetector
import org.eclipse.ui.texteditor.ITextEditor
import de.isse.kiv.resources.ResourceLookup
import de.isse.kiv.resources.ResourceLookup._
import de.isse.kiv.resources.ResourceProperties.toResourceProperties
import de.isse.kiv.ui.editors.FileEditor.openFile
import kiv.expr._
import de.isse.kiv.ui.Console
import kiv.parser.Terminal
import de.isse.kiv.refactoring.KIVRefactoring
import de.isse.kiv.project.Definition
import kiv.printer.prettyprint._
import de.isse.kiv.source.parser.SpecificationNameToken
import de.isse.kiv.source.parser.ProcedureToken
import de.isse.kiv.source.parser.OperationToken
import de.isse.kiv.source.parser.KIVToken

object Hyperlink {
  def apply(region: IRegion, text: String, label: String): Hyperlink = {
    new Hyperlink(region, text, label) {
      def open() = {}
    }
  }

  def apply(region: IRegion, text: String, label: String, openf: => Unit): Hyperlink = {
    new Hyperlink(region, text, label) {
      def open() = openf
    }
  }
}

abstract class Hyperlink(val getHyperlinkRegion: IRegion, val getHyperlinkText: String, val getTypeLabel: String) extends IHyperlink

class HyperlinkDetector(val editor: ITextEditor, refactoring: KIVRefactoring) extends IHyperlinkDetector with WithSpec {

  def defLinks(kivToken: KIVToken, region: IRegion): List[Hyperlink] = {
    val file = ResourceLookup.getIFileFromITextEditor(editor)
    if (file == null)
      return Nil

    kivToken match {
      case SpecificationNameToken(strloc) =>
        val specname = strloc.str
        val specFile = specFileOrLibrary(specname)
        val seqFile = seqFileOrLibrary(specname)
        if (specFile.fileExists && seqFile.fileExists) {
          val specLink = Hyperlink(region, s"open ${specname} specification", "", openFile(specFile))
          val seqLink = Hyperlink(region, s"open ${specname} sequents", "", openFile(seqFile))
          specLink :: seqLink :: Nil
        } else
          Nil
      case _ =>
        Definition.gotoDefinitionAction(file, kivToken) match {
          case Some(action) =>
            Hyperlink(region, s"Go to definition", "", action()) :: Nil
          case None =>
            Nil
        }
    }
  }

  def opLinks(op: Op, rectypes: List[TyCo], region: IRegion): List[IHyperlink] = op.typ match {
    case Funtype(tys, res) =>
      tys.zipWithIndex.collect {
        case (Sorttype(tyco), i) if rectypes contains tyco =>
          val pos = i // capture i statically for the closure
          val descr = "position " + pos + ": " + tyco.sortstring
          Hyperlink(region, descr, op.opstring, refactoring.addRecursiveDefinition(op, pos))
      }
    case _ =>
      Nil
  }

  def opLinks(word: String, region: IRegion): List[IHyperlink] = {
    spec match {
      case None => Nil
      case Some(spec) =>
        val allops = spec.specconsts ++ spec.specfcts ++ spec.specprds
        val matchops = allops.filter(_.opstring == word)
        val rectypes = spec.specgens.filter { _.freep }.flatMap(_.gensortlist).map { _.toSort }

        matchops.flatMap(opLinks(_, rectypes, region))
    }
  }

  /** Gets called if Ctrl is pressed and shows a list of applicable actions based on the type of the token under the mouse cursor */
  override def detectHyperlinks(textViewer: ITextViewer, _region: IRegion, canShowMultipleHyperlinks: Boolean): Array[IHyperlink] = {
    val kivToken = fileModel.tokenAt(_region.getOffset).getOrElse(return null)

    val region = new Region(kivToken.location.position, kivToken.location.length)
    val links = new ArrayBuffer[IHyperlink]

    val links1 = defLinks(kivToken, region)
    val links2 = opLinks(kivToken.stringPresentation, region)

    links ++= links1

    if (!links1.isEmpty && !links2.isEmpty)
      links += Hyperlink(region, "------------------------------------", "")

    if (!links2.isEmpty)
      links += Hyperlink(region, "Select position for recursive definition:", "")

    links ++= links2

    val array = links.toArray
    if (array.size != 0) array else null // NOTE: convention expected from eclipse's HyperlinkManager
  }
}
