package de.isse.kiv.source.analyzers

import kiv.spec._
import kiv.prog.Anydeclaration
import kiv.parser.Location

/**
 * Base class for analyzing specification/sequents files
 * Analyzers yield warnings for common problems in KIV specifications/theorems
 */
trait Analyzer {
  def checkTheorem(theorem: Theorem, spec: Spec): List[Analyzer.Warning] = Nil
  def checkDeclaration(declation: Anydeclaration, spec: Spec): List[Analyzer.Warning] = Nil
}

object Analyzer {
  /** A warning returned by an analyzer */
  case class Warning(val location: Location, val text: String) {
    override def toString = {
      text + " " + location.toString()
    }
  }

  /** The buildin analyzers */
  private val analyzers = OverwrittenVariables :: FreeVariables :: Nil

  /** Check a specification file against all analyzers */
  def checkSpecification(spec: Spec): List[Warning] = {
    val decls = spec match {
      case _: Basicspec | _: Genspec | _: Enrichedspec | _: ASMspec | _: DataASMSpec | _: Complexspec =>
        spec.decllist
      case _ =>
        Nil
    }

    val axioms = spec match {
      case _: Basicspec | _: Genspec | _: Enrichedspec | _: DataASMSpec | _: ReducedDataASMSpec | _: Complexspec =>
        spec.axiomlist
      case _ =>
        Nil
    }

    analyzers.flatMap { analyzer =>
      decls.flatMap { analyzer.checkDeclaration(_, spec) } ++
      axioms.flatMap { analyzer.checkTheorem(_, spec) }
    }
  }

  /** Check a sequents file against all analyzers */
  def checkSequents(theorems: List[Theorem]) = {
    analyzers.flatMap { analyzer =>
      theorems.flatMap { analyzer.checkTheorem(_, null) }
    }
  }
}
