package de.isse.kiv.source.analyzers

import kiv.parser._
import scala.collection.mutable.ListBuffer
import kiv.spec._
import kiv.expr.Xov
import kiv.util.ScalaExtensions._
import kiv.printer.prettyprint.xpp
import kiv.expr.Op
import kiv.signature.globalsig
import de.isse.kiv.source.analyzers.Analyzer.Warning

object FreeVariables extends Analyzer {

  override def checkTheorem(axiom: Theorem, spec: Spec): List[Warning] = {
    if (axiom.preseq == null)
      Nil
    else {
      axiom.preseq.patsuc match {
        case prefl1: PreFl1 =>
          if (prefl1.patfmalist1.size != 1)
            Nil
          else {
            val expr = prefl1.patfmalist1.head
            splitEquivalence(expr) match {
              case Some((pres, lhs, rhs)) =>
                val startvars = (lhs :: pres ++ axiom.preseq.patant.asInstanceOf[PreFl1].patfmalist1).flatMap { findStartVars }
                createWarningsForFreeVariables(rhs, startvars)
              case None =>
                Nil
            }
          }
        case _ =>
          Nil
      }
    }
  }

  private def createWarningsForFreeVariables(rightSide: PreExpr, params: List[Xov]): List[Warning] = {
    val freeVars = matchFreeVariable(rightSide, params)
    freeVars.flatMap { variable =>
      variable.location.map { location =>
        Warning(location, s"Variable ${xpp(variable.xov)} unbound by left hand side!")
      }
    }
  }

  private def matchFreeVariable(preExpr: PreExpr, params: List[Xov]): List[PreXov] = {
    val matchFV = (expr: PreExpr) => matchFreeVariable(expr, params)
    val matchMV = (expressions: List[PreExpr]) => matchFreeVariable(expressions, params)

    preExpr match {
      case preXov: PreXov =>
        getFreeVariable(preXov, params)
      case PreAp(expr, exprs) =>
        matchFV(expr) ++ matchMV(exprs)
      case PreLambda(_, vl: PreVl1, expr) =>
        matchFreeVariable(expr, params ++ findStartVars(vl.patvarlist1))
      case PreAll(_, vl: PreVl1, expr) =>
        matchFreeVariable(expr, params ++ findStartVars(vl.patvarlist1))
      case PreEx(_, vl: PreVl1, expr) =>
        matchFreeVariable(expr, params ++ findStartVars(vl.patvarlist1))
      case PreNumexpr(expr: PreExpr) =>
        matchFV(expr)
      case _ =>
        Nil
    }
  }

  private def matchFreeVariable(preExpressions: List[PreExpr], params: List[Xov]): List[PreXov] = {
    preExpressions.flatMap { matchFreeVariable(_, params) }
  }

  private def getFreeVariable(prexov: PreXov, params: List[Xov]): List[PreXov] = {
    if (!params.contains(prexov.xov))
      prexov :: Nil
    else
      Nil
  }

  private def findStartVars(preExpr: PreExpr): List[Xov] = {
    preExpr match {
      case preXov: PreXov =>
        preXov.xov :: Nil
      case _: PreOp =>
        Nil
      case PreAp(expr, exprs) =>
        findStartVars(expr) ++ findStartVars(exprs)
      case PreLambda(_, vl: PreVl1, expr) =>
        findStartVars(vl.patvarlist1) ++ findStartVars(expr)
      case PreAll(_, vl: PreVl1, expr) =>
        findStartVars(vl.patvarlist1) ++ findStartVars(expr)
      case PreEx(_, vl: PreVl1, expr) =>
        findStartVars(vl.patvarlist1) ++ findStartVars(expr)
      case PreNumexpr(expr) =>
        findStartVars(expr)
      case _ =>
        Nil
    }
  }

  private def findStartVars(listExpressions: List[PreExpr]): List[Xov] = {
    listExpressions.flatMap { expr => findStartVars(expr) }
  }

  private def splitPrecondition(preExpr: PreExpr): (List[PreExpr], PreExpr) = {
    preExpr match {
      case PreAp(preop: PreOp, exprs) if preop.op == kiv.signature.globalsig.imp_rop =>
        (exprs(0) :: Nil, exprs(1))
      case _ =>
        (Nil, preExpr)
    }
  }

  private def splitEquivalence(preExpr: PreExpr): Option[(List[PreExpr], PreExpr, PreExpr)] = {
    val (precondition, equivalence) = splitPrecondition(preExpr)
    equivalence match {
      case PreAp(preop: PreOp, exprs) if preop.op == kiv.signature.globalsig.equiv_rop =>
        Some(precondition, exprs(0), exprs(1))
      case PreAp(preop: PreOp, exprs) if preop.op == kiv.signature.globalsig.eq_rop =>
        Some(precondition, exprs(0), exprs(1))
      case _ =>
        None
    }
  }
}
