package de.isse.kiv.source.analyzers

import kiv.parser._
import scala.collection.mutable.ListBuffer
import kiv.spec._
import kiv.expr.Xov
import kiv.util.ScalaExtensions._
import kiv.printer.prettyprint.xpp
import kiv.prog.Anydeclaration
import de.isse.kiv.source.analyzers.Analyzer.Warning

object OverwrittenVariables extends Analyzer {

  override def checkDeclaration(declaration: Anydeclaration, spec: Spec): List[Warning] = {
    val overwrittenVars = matchOverwrittenVariable(declaration.preprog, declaration.declprocdecl.fpl.allparams)
    overwrittenVars.flatMap { variable =>
      variable.location.map { location =>
        Warning(location, s"Variable ${xpp(variable.xov)} should not be overwritten")
      }
    }
  }

  private def matchOverwrittenVariable(preProg: PreProg, params: List[Xov]): List[PreXov] = {
    (preProg: @unchecked) match {
      case _: NullaryProg =>
        Nil
      case _: PreSkip | _: PreAbort | _: PrePblocked | _: PreParasgmv | _: PreProgmv | _: PreParasg3
        | _: PreParasg1 | _: PreCall | _: PrePrecall | _: PreBcall | _: PreAwait | _: PreExprprog
        | _: PreAnnotation | _: PreAssert | _: PreThrow =>
        Nil
      case PreChoose(_, patchoosevl: PreVl1, patbxp, prog, prog2) =>
        filterList(params, patchoosevl.patvarlist1) ++
          matchOverwrittenVariable(prog, params ++ filterXov(patchoosevl.patvarlist1)) ++
          prog2.map { matchOverwrittenVariable(_, params ++ filterXov(patchoosevl.patvarlist1)) }.getOrElse(Nil)
      case PreVblock(_, patvdl: PreVdl1, prog) =>
        filterList(params, convertToPreExprList(patvdl.patvdecllist1)) ++
          matchOverwrittenVariable(prog, params ++ filterXov(convertToPreExprList(patvdl.patvdecllist1)))
      case PrePstar(prog, _) =>
        matchOverwrittenVariable(prog, params)
      case PreWhen(_, prog) =>
        matchOverwrittenVariable(prog, params)
      case PreAtomic(_, _, _, prog) =>
        matchOverwrittenVariable(prog, params)
      case PreComp(prog1, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ matchOverwrittenVariable(prog2, params)
      case PreRpar(prog1, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ matchOverwrittenVariable(prog2, params)
      case PreSpar(prog1, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ matchOverwrittenVariable(prog2, params)
      case PreApar(prog1, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ matchOverwrittenVariable(prog2, params)
      case PrePor(prog1, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ matchOverwrittenVariable(prog2, params)
      case PreIpar(_, prog1, _, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ matchOverwrittenVariable(prog2, params)
      case PreIparl(_, prog1, _, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ matchOverwrittenVariable(prog2, params)
      case PreIparr(_, prog1, _, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ matchOverwrittenVariable(prog2, params)
      case PreIparlb(_, prog1, _, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ matchOverwrittenVariable(prog2, params)
      case PreIparrb(_, prog1, _, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ matchOverwrittenVariable(prog2, params)
      case PreNfipar(_, prog1, _, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ matchOverwrittenVariable(prog2, params)
      case PreNfiparl(_, prog1, _, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ matchOverwrittenVariable(prog2, params)
      case PreNfiparr(_, prog1, _, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ matchOverwrittenVariable(prog2, params)
      case PreNfiparlb(_, prog1, _, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ matchOverwrittenVariable(prog2, params)
      case PreNfiparrb(_, prog1, _, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ matchOverwrittenVariable(prog2, params)
      case PreIf(_, _, prog1, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ prog2.map { matchOverwrittenVariable(_, params) }.getOrElse(Nil)
      case PreItlIf(_, _, prog1, prog2) =>
        matchOverwrittenVariable(prog1, params) ++ prog2.map { matchOverwrittenVariable(_, params) }.getOrElse(Nil)
      case PreWhile(_, _, prog) =>
        matchOverwrittenVariable(prog, params)
      case PreItlWhile(_, _, prog) =>
        matchOverwrittenVariable(prog, params)
      case PreBreak(prog, _) =>
        matchOverwrittenVariable(prog, params)
      case PreLoop(_, prog, _) =>
        matchOverwrittenVariable(prog, params)
      case PreForall(_, _, _, prog) =>
        matchOverwrittenVariable(prog, params)
      case PreTryCatch(prog, handlers) =>
        matchOverwrittenVariable(prog, params) ++ handlers.flatMap { handler => matchOverwrittenVariable(handler.prog, params) }
    }
  }

  private def filterList(params: List[Xov], patvarlist: List[PreExpr]): List[PreXov] = {
    patvarlist.filterType[PreXov].filter { prexov => params.contains { prexov.xov } }
  }

  private def convertToPreExprList(list: List[PreVdecl]): List[PreExpr] = {
    list.map {
      case PreVardecl(patvar, _) =>
        patvar
      case PreRvardecl(patvar) =>
        patvar
    }
  }

  private def filterXov(patvarlist: List[PreExpr]): List[Xov] = {
    patvarlist.filterType[PreXov].map { _.xov }
  }
}
