package de.isse.kiv.source.parser

import org.eclipse.jface.text.rules.ITokenScanner
import org.eclipse.jface.text.rules.IToken
import org.eclipse.jface.text.IDocument
import kiv.parser.Lexer
import kiv.parser.Terminals._
import de.isse.kiv.resources.ResourceLookup
import org.eclipse.jface.text.rules.Token
import org.eclipse.core.resources.IFile
import kiv.parser.Terminal
import java.io.StringReader
import kiv.parser.Terminals
import scala.collection.mutable.ListBuffer
import de.isse.kiv.resources.ResourceProperties._
import de.isse.kiv.ui.Console
import kiv.prog.NewProc
import kiv.parser.StringAndLocation
import kiv.spec._
import kiv.prog._
import kiv.parser._
import de.isse.kiv.source.analyzers._
import de.isse.kiv.source.parser._
import de.isse.kiv.source._

object KIVTokenizer {

  def parse(text: String, file: IFile) {
    val tokens = tokenize(text, Some(file), 0)
    file.fileModel.updateTokens(tokens)
  }

  def parse(text: String, file: IFile, prespec: PreSpecification) {
    val tokens = tokenize(text, Some(file), 0)
    file.fileModel.updateTokens(tokens)
  }

  def parse(text: String, file: IFile, theoremlist: List[Theorem]) {
    val tokens = tokenize(text, Some(file), 0)
    file.fileModel.updateTokens(tokens)
  }

  private def tokenize(text: String, file: Option[IFile], startPosition: Int) = {
    val projectModel = file.map { _.getProject.projectModel }
    val signatureAfter = projectModel.map { _.signatureAfter(file.get) }
    val psig = signatureAfter.getOrElse(kiv.signature.globalsig.empty_sig).toParsersig
    val parser = new kiv.parser.Parse(null, Nil, Map(), psig)
    val lexer = new Lexer(new StringReader(text), parser)
    val tokens = new ListBuffer[KIVToken]()
    val procMap = projectModel match {
      case None => Map[String, NewProc]()
      case Some(projectModel) => {
        projectModel.getProcMap(file.get)
      }
    }
    val specAnalyzer: Option[SpecMapsAnalyzer] = file.map { _.fileModel.getSpecAnalyzer }.flatten

    var sym: Terminal = null
    do {
      sym = try {
        lexer.nextToken.asInstanceOf[Terminal]
      } catch {
        case e: Exception =>
          Console.debug.println("lexing failed with exception: " + e)
          new Terminal(Terminals.EOF, 0, 0, 0, 0, null)
      }
      val offset = sym.getPosition + startPosition
      val length = sym.getLength
      val end = offset + length
      if (sym.getId() == T_PROC_SYM) {
        val str = sym.value match {
          case str: String                 => str
          case StringAndLocation(str, loc) => str
        }
        if (procMap.contains(str))
          sym = new Terminal(T_PROC, sym.getLine(), sym.getColumn(), sym.getPosition(), sym.getLength(), str)
      }

      if (sym.getId() == T_OLDXOV || sym.getId == T_PRIMEDXOV) {
        // Split into two tokens
        val othersym = sym.getId match {
          case T_OLDXOV => new Terminal(T_OLDXOV, sym.getLine(), sym.getColumn(), sym.getPosition() + sym.getLength(), 1, "`")
          case T_PRIMEDXOV =>
            val value = if (sym.value.toString().endsWith("''")) "''" else "'"
            new Terminal(T_PRIMEDXOV, sym.getLine(), sym.getColumn(), sym.getPosition() + sym.getLength(), value.length, value)
        }
        val xovsym = new Terminal(T_XOV, sym.getLine(), sym.getColumn(), sym.getPosition(), sym.getLength() - othersym.getLength, sym.value)
        val otherstr = othersym.value.asInstanceOf[String]

        specAnalyzer match {
          case Some(specAnalyzer) => {
            specAnalyzer.prexovMap.get(getRange(xovsym, startPosition)) match {
              case Some(prexov) => { tokens += new VariableToken(prexov); }
              case None         => tokens += new RawToken(xovsym, text.substring(offset - startPosition, (end - startPosition) - otherstr.length), offset until (end - otherstr.length))
            }
          }
          case None => tokens += new RawToken(xovsym, text.substring(offset - startPosition, (end - startPosition) - otherstr.length), offset until (end - otherstr.length))
        }
        tokens += new RawToken(othersym, otherstr, (end - otherstr.length) until end)
      } else {
        specAnalyzer match {
          case Some(specAnalyzer) =>
            val range = getRange(sym, startPosition)
            specAnalyzer.specificationMap.get(range) match {
              case Some(strloc) =>
                tokens += new SpecificationNameToken(strloc)
              case None =>
                specAnalyzer.typesMap.get(range) match {
                  case Some(tyco) =>
                    tokens += new SortDefinitionToken(tyco)
                  case None =>
                    specAnalyzer.prexovMap.get(range) match {
                      case Some(prexov) =>
                        tokens += new VariableToken(prexov)
                      case None =>
                        specAnalyzer.procMap.get(range) match {
                          case Some(preproc) =>
                            tokens += new ProcedureToken(preproc)
                          case None =>
                            specAnalyzer.preopMap.get(range) match {
                              case Some(preop) =>
                                tokens += OperationToken(preop)
                              case None =>
                                specAnalyzer.theoremMap.get(range) match {
                                  case Some(strloc) =>
                                    tokens += TheoremNameToken(strloc)
                                  case None =>
                                    specAnalyzer.pretycoMap.get(range) match {
                                      case Some(pretyco) =>
                                        tokens += SortToken(pretyco)
                                      case None =>
                                        specAnalyzer.opdefMap.get(range) match {
                                          case Some(opdef) =>
                                            tokens += OperationDefinitionToken(opdef)
                                          case None =>
                                            tokens += new RawToken(sym, text.substring(offset - startPosition, end - startPosition), offset until end)
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
          case None =>
            tokens += new RawToken(sym, text.substring(offset - startPosition, end - startPosition), offset until end)
        }
      }
    } while (sym.getId != Terminals.EOF)
    tokens.toList
  }

  private def getLocation(terminal: Terminal) = {
    new Location(terminal.getLine, terminal.getColumn, terminal.getPosition, terminal.getLength)
  }

  private def getRange(terminal: Terminal, position: Int): Range = {
    val start = position + terminal.getPosition
    val end = start + terminal.getLength
    start until end
  }

}

class KIVTokenizer extends ITokenScanner {
  var doc: IDocument = _
  var rangeLastToken: Range = _
  var tokens: List[KIVToken] = Nil
  var cache: List[(IToken, Range)] = Nil

  def getTokenOffset(): Int = rangeLastToken.start
  def getTokenLength(): Int = rangeLastToken.length

  def setRange(doc: IDocument, start: Int, length: Int) {
    this.doc = doc
    this.rangeLastToken = 0 until 0

    val file = ResourceLookup.iDocumentToIFile(doc)
    val filemodel = file.map { _.fileModel }

    // HACK: we tokenize the full file, otherwise it does not work for unknown reasons
    this.tokens = KIVTokenizer.tokenize(doc.get /*.get(start, length)*/ , file, 0) //start)

    //tokens.foreach { x => println(x.stringPresentation+" "+x.location) }
    filemodel match {
      case Some(filemodel) =>
        filemodel.updateTokens(tokens)
      case None =>
    }
    this.cache = Nil
  }

  def nextToken(): IToken = {
    // cache.foreach(x  => println(x._2))
    if (cache.isEmpty) feed()
    assert(!cache.isEmpty)

    val (tok, range) = cache.head
    this.cache = cache.tail
    this.rangeLastToken = range

    if (tok.isEOF())
      assert(tokens.isEmpty && cache.isEmpty)

    tok
  }

  def firstWhitespace(s: String): Int = {
    val pos = s.indexWhere(!_.isWhitespace)
    if (pos < 0) s.length else pos
  }

  def feed() {
    val headToken: KIVToken = tokens.head
    tokens = tokens.tail

    val tok = headToken.eclipseToken
    cache = (tok, (headToken.location.position until (headToken.location.position + headToken.location.length))) :: cache

    val prevOffset = this.rangeLastToken.end
    val spaceLength = headToken.location.position - prevOffset
    if (spaceLength > 0) {
      val text = doc.get(prevOffset, spaceLength)
      val commentStart = prevOffset + firstWhitespace(text)
      val commentEnd = commentStart + text.trim.length

      /* tokens are pushed in right-to-left order */
      if (commentEnd < headToken.location.position) {
        cache = (Token.WHITESPACE, commentEnd until headToken.location.position) :: cache
      }

      if (commentStart < commentEnd) {
        cache = (TokenMap.COMMENT, commentStart until commentEnd) :: cache
      }
      // why was the offset shifted by +/- 1 ???
      // this produced e.g. strange highlighting behavior when several program constructs appeared directly next to each other
      // UPDATE: it seems it isn't allowed that the first eclipse token is a WHITESPACE token, so simply omit it?
      if (prevOffset < commentStart && prevOffset != 0) {
        cache = (Token.WHITESPACE, prevOffset until (commentStart)) :: cache
      }
    }
    /*cache.foreach {
      case first =>
        cache.foreach {
          case second if first != second =>
            if (overlaps(first._2, second._2))
              de.isse.kiv.ui.Console.error.println("bla "+first+" "+second)
          case _ =>
        }
    }*/
  }

  def overlaps(r0: Range, r1: Range): Boolean = {
    assert(!r0.isInclusive && !r1.isInclusive)
    (r0.start <= r1.start && r0.end > r1.start) || (r1.start <= r0.start && r1.end > r0.start)
  }
}
