package de.isse.kiv.source.parser

import kiv.spec._
import kiv.prog._
import kiv.parser._
import kiv.expr._
import kiv.util.ScalaExtensions._

class PreSpecificationAnalyzer(val prespec: PreSpecification, override val procnameMap: Map[String, AnyProc]) extends SpecMapsAnalyzer {

  val (listExpr, listDecl) = getPre(prespec)

  override lazy val specificationMap: Map[Range, StringAndLocation] = {
    prespec.specifications.map { specloc =>
      (specloc.strloc.loc.toRange, specloc.strloc)
    }.toMap
  }

  override lazy val typesMap: Map[Range, TyCodef] = {
    prespec.toplevelTypeDefs.map { tycodef =>
      (tycodef.symloc.loc.toRange, tycodef)
    }.toMap
  }

  override lazy val pretycoMap: Map[Range, PreTyCo] = {
    // could contain dummy locations (0,0,0,0) -> possible conflict with EOF token
    // TODO can we avoid using dummy locations?
    val premap = prespec.toplevelTypes.flatMap { (pretype: PreType) =>
      pretype.pretycos map { pretyco => (pretyco.pretycosymloc.loc.toRange, pretyco) }
    }.toMap
    premap filterNot (_._1.length == 0)
  }

  override lazy val opdefMap: Map[Range, Opdef] = {
    // could contain dummy locations (0,0,0,0) -> possible conflict with EOF token
    // TODO can we avoid using dummy locations?
    val premap = prespec.toplevelOpdefs.map { opdef =>
      (opdef.opdefsymloc.loc.toRange, opdef)
    }.toMap
    premap filterNot (_._1.length == 0)
  }

  override lazy val theoremMap: Map[Range, StringAndLocation] = {
    val pairs = prespec.theorems.map { theorem =>
      (theorem.location.toRange, StringAndLocation(theorem.theoremname, theorem.location))
    } ++
    prespec.declarations.map { decl =>
      val strloc = decl.nameandloc
      (strloc.loc.toRange -> strloc)
    }
    pairs.toMap[Range, StringAndLocation]
  }

  override lazy val prexovMap: Map[Range, PreAnyXov] = {
    getAllPreXov(listExpr, listDecl).map { prexov =>
      (prexov.location.get.toRange -> prexov)
    }.toMap
  }

  override lazy val procMap: Map[Range, PreProc] = {
    (getAllPreProcDecl(listDecl) ++ getAllPotentialPreProc(prespec.potentialProcedures)).flatMap {
      preproc =>
        preproc.location match {
          case Some(location) =>
            Some(location.toRange -> preproc)
          case None =>
            de.isse.kiv.ui.Console.error.println("Location from anyproc: " + preproc + " is missing")
            None
        }
    }.toMap
  }

  override lazy val preopMap: Map[Range, PreOp] = {
    getAllPreOp(listExpr, listDecl).flatMap {
      preop =>
        preop.location match {
          case Some(location) =>
            Some(location.toRange -> preop)
          case None =>
            // de.isse.kiv.ui.Console.error.println("Location from preop: "+preop+" is missing!")
            None
        }
    }.toMap
  }

  private def getPre(prespec: PreSpecification): (List[PreExpr], List[PreDeclaration]) = {
    (getAllPreExpr(prespec), getAllPredeclarations(prespec))
  }

  private def getAllPreExpr(prespec: PreSpecification) = {
    prespec.toplevelExprs ++
      prespec.theorems.flatMap { theorem => getAllPreExprFromTheorem(theorem) }
  }

  private def getAllPreExprFromTheorem(theorem: Theorem): List[PreExpr] = {
    if (theorem.preseq == null)
      Nil
    else theorem.preseq.patsuc match {
      case prefl1: PreFl1 => prefl1.patfmalist1 ++ theorem.preseq.patant.asInstanceOf[PreFl1].patfmalist1;
      case prefl3: PreFl3 => prefl3.patfmalist1 ++ prefl3.patfmalist2 ++ (theorem.preseq.patant match {
        case prefl1: PreFl1 => prefl1.patfmalist1
        case prefl3: PreFl3 => prefl3.patfmalist1 ++ prefl3.patfmalist2
        case _              => Nil
      })
      case _ => Nil
    }
  }

  private def getAllPredeclarations(prespec: PreSpecification) = {
    prespec.declarations
  }

  private def getAllPreXov(preexprList: List[PreExpr], predeclarationList: List[PreDeclaration]): List[PreAnyXov] = {
    getAllPreXov(preexprList) ++ getAllPreXovDecl(predeclarationList)
  }

  private def getAllPreXovDecl(predeclarationList: List[PreDeclaration]): List[PreAnyXov] = {
    predeclarationList.flatMap { predeclaration => getAllPreXovDecl(predeclaration) }
  }

  private def getAllPreXovDecl(predeclaration: PreDeclaration): List[PreAnyXov] = {
    predeclaration match {
      case PreOperationDeclaration(nameandloc: StringAndLocation, procsym: SymbolAndLocation,
        fpl: PreFpl, contract: Option[PreContract],
        preprog: PreProg, prog: Prog) =>
        getAllPreXov(fpl.outparams) ++ getAllPreXov(fpl.valueparams) ++ getAllPreXov(fpl.varparams) ++
          getAllPreXov(preprog) ++ (contract match {
            case Some(contract: PreContract) =>
              val run_pre_xov = contract.run match {
                case Some((pre_run_fma, pre_run_ident)) =>
                  getAllPreXov(pre_run_fma) ++ getAllPreXov(pre_run_ident)
                case _ =>
                  Nil
              }
              getAllPreXov(contract.pre) ++ getAllPreXov(contract.guar) ++ getAllPreXov(contract.post) ++ run_pre_xov
            case None =>
              Nil
          })
    }
  }

  private def getAllPreProcDecl(predeclarationList: List[PreDeclaration]): List[PreProc] = {
    predeclarationList.flatMap { predeclaration => getAllPreProc(predeclaration) }
  }

  private def getAllPreProc(predeclaration: PreDeclaration): List[PreProc] = {
    predeclaration match {
      case PreOperationDeclaration(nameandloc: StringAndLocation, procsym: SymbolAndLocation,
        fpl: PreFpl, contract: Option[PreContract],
        preprog: PreProg, prog: Prog) =>
        getAllPreProc(preprog) ++ checkNameAndProcSym(nameandloc, procsym)
    }
  }

  private def getAllPotentialPreProc(list: List[SymbolAndLocation]): List[PreProc] = {
    list.flatMap { symbolloc => checkSymbolLocation(symbolloc) }
  }

  private def checkNameAndProcSym(nameandloc: StringAndLocation, procsym: SymbolAndLocation): List[PreProc] = {
    (procnameMap.get(nameandloc.str) match {
      case Some(proc) =>
        val preproc = PreProc(proc, nameandloc.loc)
        preproc :: Nil
      case None => Nil
    }) ++
      checkSymbolLocation(procsym)
  }

  private def checkSymbolLocation(symbolloc: SymbolAndLocation): List[PreProc] = {
    (procnameMap.get(symbolloc.sym.name) match {
      case Some(proc) =>
        val preproc = PreProc(proc, symbolloc.loc)
        preproc :: Nil
      case None => Nil
    })
  }

  private def getAllPreOp(preexprList: List[PreExpr], predeclarationList: List[PreDeclaration]): List[PreOp] = {
    getAllPreOp(preexprList) ++ getAllPreOpDecl(predeclarationList)
  }

  private def getAllPreOpDecl(predeclarationList: List[PreDeclaration]): List[PreOp] = {
    predeclarationList.flatMap { predeclaration => getAllPreOpDecl(predeclaration) }
  }

  private def getAllPreOpDecl(predeclaration: PreDeclaration): List[PreOp] = {
    predeclaration match {
      case PreOperationDeclaration(nameandloc: StringAndLocation, procsym: SymbolAndLocation,
        fpl: PreFpl, contract: Option[PreContract],
        preprog: PreProg, prog: Prog) =>
        getAllPreOp(fpl.outparams) ++ getAllPreOp(fpl.valueparams) ++ getAllPreOp(fpl.varparams) ++
          getAllPreOp(preprog) ++ (contract match {
            case Some(contract: PreContract) =>
              val run_pre_op = contract.run match {
                case Some((pre_run_fma, pre_run_ident)) =>
                  getAllPreOp(pre_run_fma) ++ getAllPreOp(pre_run_ident)
                case _ =>
                  Nil
              }
              getAllPreOp(contract.pre) ++ getAllPreOp(contract.guar) ++ getAllPreOp(contract.post) ++ run_pre_op
            case None =>
              Nil
          })
    }
  }
}
