package de.isse.kiv.source.parser

import kiv.spec._
import kiv.prog._
import kiv.parser._
import kiv.expr.Xov
import kiv.util.ScalaExtensions._
import kiv.expr.Op
import kiv.signature.Sigentry

trait SpecMapsAnalyzer {
  def specificationMap: Map[Range, StringAndLocation] = Map()
  def typesMap: Map[Range, TyCodef]
  def prexovMap: Map[Range, PreAnyXov]
  def procMap: Map[Range, PreProc]
  def preopMap: Map[Range, PreOp]
  def pretycoMap: Map[Range, PreTyCo]
  def opdefMap: Map[Range, Opdef]
  def theoremMap: Map[Range, StringAndLocation] = Map()
  def procnameMap: Map[String, AnyProc]

  def getLocationsFromSigentry(signature: Sigentry): List[Location] = {
    signature match {
      case xov: Xov         => getLocationsFromXov(xov)
      case anyproc: AnyProc => getLocationsFromAnyProc(anyproc)
      case op: Op           => getLocationsFromOp(op)
      case _                => Nil
    }
  }

  private def getLocationsFromXov(xov: Xov): List[Location] = {
    prexovMap.filter(x => (x._2 match {
      case prexov: PreXov => prexov.xov == xov
      case _              => false
    })).map(x => x._2.location.get).toList
  }

  private def getLocationsFromAnyProc(anyproc: AnyProc) = {
    procMap.filter(x => x._2.proc == anyproc).map(x => x._2.location.get).toList
  }

  private def getLocationsFromOp(op: Op) = {
    preopMap.filter(x => x._2.op == op).map(x => x._2.location.get).toList
  }

  def getAllPreXov(preexprList: List[PreExpr]): List[PreAnyXov] = {
    preexprList.flatMap { preexpr => getAllPreXov(preexpr) }
  }

  def getAllPreXov(preExpr: PreExpr): List[PreAnyXov] = {
    preExpr match {
      case preXov: PreXov         => List(preXov)
      case poly: PrePolyXov       => List(poly)
      case unary: UnaryExpr       => getAllPreXov(unary.fma)
      case primedXov: PrimedXov   => getAllPreXov(primedXov.fma)
      case binaryExpr: BinaryExpr => getAllPreXov(binaryExpr.fma1) ++ getAllPreXov(binaryExpr.fma2)
      case progWithPostCondition: ProgWithPostCondition => getAllPreXov(progWithPostCondition.fma) ++
        getAllPreXov(progWithPostCondition.prog)
      case PreRgbox(_, vl, rely, guar, inv, prog, fma, _) =>
        getAllPreXov(fma) ++ getAllPreXov(guar) ++ getAllPreXov(inv) ++
          getAllPreXov(prog) ++ getAllPreXov(rely) ++
          (vl match {
            case prevl1: PreVl1 => getAllPreXov(prevl1.patvarlist1)
            case _              => Nil
          })
      case PreRgdia(_, vl, rely, guar, inv, run, prog, fma, _) =>
        getAllPreXov(fma) ++ getAllPreXov(guar) ++ getAllPreXov(inv) ++ getAllPreXov(run)
        getAllPreXov(prog) ++ getAllPreXov(rely) ++
          (vl match {
            case prevl1: PreVl1 => getAllPreXov(prevl1.patvarlist1)
            case _              => Nil
          })
      case PreAp(expr, listOfExpressions) => getAllPreXov(expr) ++ getAllPreXov(listOfExpressions)
      case PreLambda(_, vl: PreVl1, expr: PreExpr) => getAllPreXov(expr) ++ getAllPreXov(vl.patvarlist1)
      case PreAll(_, vl: PreVl1, expr: PreExpr) => getAllPreXov(expr) ++ getAllPreXov(vl.patvarlist1)
      case PreEx(_, vl: PreVl1, expr: PreExpr) => getAllPreXov(expr) ++ getAllPreXov(vl.patvarlist1)
      case PreNumexpr(expr: PreExpr) => getAllPreXov(expr)
      case PreVarprogexpr(_, vl: PreVl1, prog: PreProg, _) => getAllPreXov(vl.patvarlist1) ++ getAllPreXov(prog)
      case _ => Nil
    }
  }

  def getAllPreXov(preProg: PreProg): List[PreAnyXov] = {
    preProg match {
      case unaryProg: UnaryProg           => getAllPreXov(unaryProg.prog)
      case PreAtomic(_, _, expr, prog)    => getAllPreXov(expr) ++ getAllPreXov(prog)
      case binaryProg: BinaryProg         => getAllPreXov(binaryProg.prog1) ++ getAllPreXov(binaryProg.prog2)
      case interleaving: Interleaving     => getAllPreXov(interleaving.prog1) ++ getAllPreXov(interleaving.prog2)
      case ifConstruct: IfConstruct       => getAllPreXov(ifConstruct.bxp) ++ getAllPreXov(ifConstruct.prog1) ++ ifConstruct.prog2.map { getAllPreXov }.getOrElse(Nil)
      case whileConstruct: WhileConstruct => getAllPreXov(whileConstruct.bxp) ++ getAllPreXov(whileConstruct.prog)
      case PreLoop(_, expr, prog)         => getAllPreXov(expr) ++ getAllPreXov(prog)
      case PreChoose(_, prevl: PreVl1, expr: PreExpr, prog: PreProg, prog2) => getAllPreXov(prevl.patvarlist1) ++ getAllPreXov(expr) ++
        getAllPreXov(prog) ++ prog2.map { getAllPreXov }.getOrElse(Nil)
      case PreForall(_, prevl: PreVl1, expr: PreExpr, prog: PreProg) => getAllPreXov(prevl.patvarlist1) ++ getAllPreXov(expr) ++ getAllPreXov(prog)
      case PreVblock(_, patvdl: PreVdl1, prog: PreProg)              => getAllPreXov(convertToPreExprList(patvdl.patvdecllist1)) ++ getAllPreXov(prog)
      case PreCall(_, preapl)                                        => getAllPreXov(preapl.return_parameters) ++ getAllPreXov(preapl.value_parameters) ++ getAllPreXov(preapl.variable_parameters)
      case PrePrecall(_, preapl)                                     => getAllPreXov(preapl.return_parameters) ++ getAllPreXov(preapl.value_parameters) ++ getAllPreXov(preapl.variable_parameters)
      case PreBcall(_, preapl, expr) => getAllPreXov(preapl.return_parameters) ++ getAllPreXov(preapl.value_parameters) ++
        getAllPreXov(preapl.variable_parameters) ++ getAllPreXov(expr)
      case PreParasg1(patassignlist1: List[PreAssign]) => getAllPreXov(convertToPreExprList2(patassignlist1))
      case PreParasg3(patassignlist1: List[PreAssign], _, patassignlist2: List[PreAssign]) => getAllPreXov(convertToPreExprList2(patassignlist1)) ++
        getAllPreXov(convertToPreExprList2(patassignlist2))
      case PreAssert(_, expr)           => getAllPreXov(expr)
      case PreAwait(_, expr)            => getAllPreXov(expr)
      case PreExprprog(_, expr, _)      => getAllPreXov(expr)
      case PreAnnotation(assertionlist) => getAllPreXov(getAllPreExprFromAssertionList(assertionlist))
      case _                            => Nil
    }
  }

  private def getAllPreExprFromAssertionList(assertionlist: List[PreAssertion]): List[PreExpr] = {
    assertionlist.flatMap { assertion => getAllPreExpr(assertion) }
  }

  private def getAllPreExpr(assertion: PreAssertion): List[PreExpr] = {
    assertion match {
      case preContractAssert: PreContractassert => preContractAssert.subst.terms ++ preContractAssert.subst.vars
      case PreWfinvassert(_, invariant: PreExpr, wfbound: PreExpr) => List(invariant, wfbound)
      case PreWfassert(_, wfbound: PreExpr) => List(wfbound)
      case PreStructassert(_, structbound: PreExpr) => List(structbound)
      case PreInvassert(_, invariant: PreExpr) => List(invariant)
      case PreCutassert(_, cutfma: PreExpr) => List(cutfma)
      case _ => Nil
    }
  }

  def getAllPreProc(expressionList: List[PreExpr]): List[PreProc] = {
    expressionList.flatMap { expr => getAllPreProc(expr) }
  }

  def getAllPreProc(preExpr: PreExpr): List[PreProc] = {
    preExpr match {
      case progWithPostCondition: ProgWithPostCondition => getAllPreProc(progWithPostCondition.prog)
      case rgbox: PreRgbox => getAllPreProc(rgbox.prog)
      case rgdia: PreRgdia => getAllPreProc(rgdia.prog)
      case prevarprog: PreVarprogexpr => getAllPreProc(prevarprog.prog)
      case _ => Nil
    }
  }

  def getAllPreProc(preProg: PreProg): List[PreProc] = {
    preProg match {
      case precall: PreCall => List(precall.proc) ++ getAllPreProc(precall.patapl.return_parameters) ++
        getAllPreProc(precall.patapl.value_parameters) ++ getAllPreProc(precall.patapl.variable_parameters)
      case preprecall: PrePrecall => (procnameMap.get(preprecall.procsym.sym.name) match {
        case Some(proc) =>
          val preproc = PreProc(proc, preprecall.procsym.loc)
          preproc :: Nil
        case None => Nil
      }) ++ getAllPreProc(preprecall.apl.return_parameters) ++ getAllPreProc(preprecall.apl.value_parameters) ++
        getAllPreProc(preprecall.apl.variable_parameters)
      case prebcall: PreBcall => List(prebcall.proc) ++ getAllPreProc(prebcall.patcxp) ++
        getAllPreProc(prebcall.patapl.return_parameters) ++ getAllPreProc(prebcall.patapl.value_parameters) ++
        getAllPreProc(prebcall.patapl.variable_parameters) ++ getAllPreProc(prebcall.patcxp)
      case unaryProg: UnaryProg           => getAllPreProc(unaryProg.prog)
      case PreAtomic(_, _, expr, prog)    => getAllPreProc(expr) ++ getAllPreProc(prog)
      case binaryProg: BinaryProg         => getAllPreProc(binaryProg.prog1) ++ getAllPreProc(binaryProg.prog2)
      case interleaving: Interleaving     => getAllPreProc(interleaving.prog1) ++ getAllPreProc(interleaving.prog2)
      case ifConstruct: IfConstruct       => getAllPreProc(ifConstruct.prog1) ++ ifConstruct.prog2.map { getAllPreProc }.getOrElse(Nil)
      case whileConstruct: WhileConstruct => getAllPreProc(whileConstruct.prog)
      case PreLoop(_, expr, prog)         => getAllPreProc(expr) ++ getAllPreProc(prog)
      case PreChoose(_, prevl: PreVl1, expr: PreExpr, prog: PreProg, prog2) => getAllPreProc(prevl.patvarlist1) ++ getAllPreProc(expr) ++
        getAllPreProc(prog) ++ prog2.map { getAllPreProc }.getOrElse(Nil)
      case PreForall(_, prevl: PreVl1, expr: PreExpr, prog: PreProg) => getAllPreProc(prevl.patvarlist1) ++ getAllPreProc(expr) ++ getAllPreProc(prog)
      case PreVblock(_, patvdl: PreVdl1, prog: PreProg)              => getAllPreProc(convertToPreExprList(patvdl.patvdecllist1)) ++ getAllPreProc(prog)
      case PreParasg1(patassignlist1: List[PreAssign])               => getAllPreProc(convertToPreExprList2(patassignlist1))
      case PreParasg3(patassignlist1: List[PreAssign], _, patassignlist2: List[PreAssign]) => getAllPreProc(convertToPreExprList2(patassignlist1)) ++
        getAllPreProc(convertToPreExprList2(patassignlist2))
      case PreAssert(_, expr)           => getAllPreProc(expr)
      case PreAwait(_, expr)            => getAllPreProc(expr)
      case PreExprprog(_, expr, _)      => getAllPreProc(expr)
      case PreAnnotation(assertionlist) => getAllPreProc(getAllPreExprFromAssertionList(assertionlist))
      case _                            => Nil
    }
  }

  def getAllPreOp(preexprList: List[PreExpr]): List[PreOp] = {
    preexprList.flatMap { preexpr => getAllPreOp(preexpr) }
  }

  def getAllPreOp(expr: PreExpr): List[PreOp] = {
    expr match {
      case preOp: PreOp           => List(preOp)
      case unary: UnaryExpr       => getAllPreOp(unary.fma)
      case primedXov: PrimedXov   => getAllPreOp(primedXov.fma)
      case binaryExpr: BinaryExpr => getAllPreOp(binaryExpr.fma1) ++ getAllPreOp(binaryExpr.fma2)
      case progWithPostCondition: ProgWithPostCondition => getAllPreOp(progWithPostCondition.fma) ++
        getAllPreOp(progWithPostCondition.prog)
      case PreRgbox(_, vl, rely, guar, inv, prog, fma, _) =>
        getAllPreOp(fma) ++ getAllPreOp(guar) ++ getAllPreOp(inv) ++
          getAllPreOp(prog) ++ getAllPreOp(rely) ++
          (vl match {
            case prevl1: PreVl1 => getAllPreOp(prevl1.patvarlist1)
            case _              => Nil
          })
      case PreRgdia(_, vl, rely, guar, inv, run, prog, fma, _) =>
        getAllPreOp(fma) ++ getAllPreOp(guar) ++ getAllPreOp(inv) ++ getAllPreOp(run) ++
          getAllPreOp(prog) ++ getAllPreOp(rely) ++
          (vl match {
            case prevl1: PreVl1 => getAllPreOp(prevl1.patvarlist1)
            case _              => Nil
          })
      case PreAp(expr, listOfExpressions) => getAllPreOp(expr) ++ getAllPreOp(listOfExpressions)
      case PreLambda(_, vl: PreVl1, expr: PreExpr) => getAllPreOp(expr) ++ getAllPreOp(vl.patvarlist1)
      case PreAll(_, vl: PreVl1, expr: PreExpr) => getAllPreOp(expr) ++ getAllPreOp(vl.patvarlist1)
      case PreEx(_, vl: PreVl1, expr: PreExpr) => getAllPreOp(expr) ++ getAllPreOp(vl.patvarlist1)
      case PreNumexpr(expr: PreExpr) => getAllPreOp(expr)
      case PreVarprogexpr(_, vl: PreVl1, prog: PreProg, _) => getAllPreOp(vl.patvarlist1) ++ getAllPreOp(prog)
      case _ => Nil
    }
  }

  def getAllPreOp(preProg: PreProg): List[PreOp] = {
    preProg match {
      case unaryProg: UnaryProg           => getAllPreOp(unaryProg.prog)
      case PreAtomic(_, _, expr, prog)    => getAllPreOp(expr) ++ getAllPreOp(prog)
      case binaryProg: BinaryProg         => getAllPreOp(binaryProg.prog1) ++ getAllPreOp(binaryProg.prog2)
      case interleaving: Interleaving     => getAllPreOp(interleaving.prog1) ++ getAllPreOp(interleaving.prog2)
      case ifConstruct: IfConstruct       => getAllPreOp(ifConstruct.bxp) ++ getAllPreOp(ifConstruct.prog1) ++ ifConstruct.prog2.map { getAllPreOp }.getOrElse(Nil)
      case whileConstruct: WhileConstruct => getAllPreOp(whileConstruct.bxp) ++ getAllPreOp(whileConstruct.prog)
      case PreLoop(_, expr, prog)         => getAllPreOp(expr) ++ getAllPreOp(prog)
      case PreChoose(_, prevl: PreVl1, expr: PreExpr, prog: PreProg, prog2) => getAllPreOp(prevl.patvarlist1) ++ getAllPreOp(expr) ++
        getAllPreOp(prog) ++ prog2.map { getAllPreOp }.getOrElse(Nil)
      case PreForall(_, prevl: PreVl1, expr: PreExpr, prog: PreProg) => getAllPreOp(prevl.patvarlist1) ++ getAllPreOp(expr) ++ getAllPreOp(prog)
      case PreVblock(_, patvdl: PreVdl1, prog: PreProg)              => getAllPreOp(convertToPreExprList(patvdl.patvdecllist1)) ++ getAllPreOp(prog)
      case PreCall(_, preapl)                                        => getAllPreOp(preapl.return_parameters) ++ getAllPreOp(preapl.value_parameters) ++ getAllPreOp(preapl.variable_parameters)
      case PrePrecall(_, preapl)                                     => getAllPreOp(preapl.return_parameters) ++ getAllPreOp(preapl.value_parameters) ++ getAllPreOp(preapl.variable_parameters)
      case PreBcall(_, preapl, expr) => getAllPreOp(preapl.return_parameters) ++ getAllPreOp(preapl.value_parameters) ++
        getAllPreOp(preapl.variable_parameters) ++ getAllPreOp(expr)
      case PreParasg1(patassignlist1: List[PreAssign]) => getAllPreOp(convertToPreExprList2(patassignlist1))
      case PreParasg3(patassignlist1: List[PreAssign], _, patassignlist2: List[PreAssign]) => getAllPreOp(convertToPreExprList2(patassignlist1)) ++
        getAllPreOp(convertToPreExprList2(patassignlist2))
      case PreAssert(_, expr)               => getAllPreOp(expr)
      case PreAwait(_, expr)                => getAllPreOp(expr)
      case PreExprprog(_, expr: PreExpr, _) => getAllPreOp(expr)
      case PreAnnotation(assertionlist)     => getAllPreOp(getAllPreExprFromAssertionList(assertionlist))
      case _                                => Nil
    }
  }

  def convertToPreExprList2(list: List[PreAssign]): List[PreExpr] = {
    list.flatMap {
      case PreAsg(patvar, patterm) =>
        List(patvar, patterm)
      case PreRasg(patvar, _) =>
        List(patvar)
      case PreCasg(patvar, patterm) =>
        List(patvar, patterm)
    }
  }

  def convertToPreExprList(list: List[PreVdecl]): List[PreExpr] = {
    list.flatMap {
      case PreVardecl(patvar, patterm) =>
        List(patvar, patterm)
      case PreRvardecl(patvar) =>
        List(patvar)
    }
  }
}
