package de.isse.kiv.ui

import java.io.IOException
import java.io.OutputStream
import org.eclipse.jface.preference.IPreferenceStore
import org.eclipse.jface.util.IPropertyChangeListener
import org.eclipse.jface.util.PropertyChangeEvent
import org.eclipse.swt.graphics.Color
import org.eclipse.ui.IWorkbenchPage
import org.eclipse.ui.PartInitException
import org.eclipse.ui.console.ConsolePlugin
import org.eclipse.ui.console.IConsole
import org.eclipse.ui.console.IConsoleConstants
import org.eclipse.ui.console.IConsoleManager
import org.eclipse.ui.console.IConsoleView
import org.eclipse.ui.console.MessageConsole
import org.eclipse.ui.console.MessageConsoleStream
import de.isse.kiv.KIVPlugin
import de.isse.kiv.Settings
import de.isse.kiv.resources.Colors
import de.isse.kiv.resources.ResourceLookup
import scala.collection.JavaConversions._
import java.io.PrintWriter

object Console {
  abstract class Output extends OutputStream {
    def println(args: AnyRef*): Unit

    def printShort(e: Exception) {
      val trace = e.getStackTrace
      val message = e.getMessage
      if (message != null)
        println(message)
      else
        println("exception without a message")
      if (! trace.isEmpty) println(trace.head)
    }

    def print(e: Throwable) {
      val printWriter = new PrintWriter(this)
      e.printStackTrace(printWriter)
      printWriter.flush()
    }
  }

  class WrappedMessageConsoleOutputStream(name: String, console: MessageConsole, color: Color) extends Output with IPropertyChangeListener {
    val preferences = KIVPlugin.getDefault.getPreferenceStore
    preferences.addPropertyChangeListener(this)

    val stream: MessageConsoleStream = console.newMessageStream()
    var enabled: Boolean = preferences.getBoolean(name)

    this.stream.setColor(color)

    def println(args: AnyRef*) {
      if (!enabled) return
      var first = true
      for (a <- args) {
        if (!first) stream.print(" ")
        stream.print(a.toString)
        first = false
      }
      stream.println()
    }

    override def propertyChange(event: PropertyChangeEvent) {
      if (event.getProperty == name) {
        val value = event.getNewValue
        enabled = (value.isInstanceOf[java.lang.Boolean]) && value.asInstanceOf[java.lang.Boolean]
      }
    }

    def write(b: Int) {
      if (enabled) stream.write(b)
    }

    override def write(b: Array[Byte], off: Int, len: Int) {
      if (enabled) stream.write(b, off, len)
    }
  }

  val CONSOLE_NAME = "KIV Console"

  val out: Output = newStream(Settings.CONSOLE_OUT, Colors.BLACK)
  val debug: Output = newStream(Settings.CONSOLE_DEBUG, Colors.GREEN)
  val error: Output = newStream(Settings.CONSOLE_ERROR, Colors.RED)

  private def findConsole(name: String): MessageConsole = {
    val plugin = ConsolePlugin.getDefault
    val manager = plugin.getConsoleManager
    val existing = manager.getConsoles
    existing.collectFirst {
      case mc: MessageConsole if mc.getName == name =>
        return mc
    }
    val mc = new MessageConsole(name, null)
    manager.addConsoles(Array(mc))
    mc
  }

  def showConsole() {
    try {
      val mc = findConsole(CONSOLE_NAME)
      val page = ResourceLookup.activePage
      val id = IConsoleConstants.ID_CONSOLE_VIEW
      val view = page.showView(id).asInstanceOf[IConsoleView]
      view.display(mc)
    } catch {
      case e: PartInitException =>
    }
  }

  private def newStream(name: String, color: Color): Output = {
    val mc = findConsole(CONSOLE_NAME)
    new WrappedMessageConsoleOutputStream(name, mc, color)
  }
}
