package de.isse.kiv.ui.editors

import org.eclipse.core.resources.IFile
import org.eclipse.jface.resource.FontRegistry
import org.eclipse.jface.text.ITextSelection
import org.eclipse.jface.text.source.DefaultCharacterPairMatcher
import org.eclipse.jface.util.IPropertyChangeListener
import org.eclipse.jface.util.PropertyChangeEvent
import org.eclipse.swt.widgets.Composite
import org.eclipse.swt.widgets.Display
import org.eclipse.ui.IEditorInput
import org.eclipse.ui.IPageLayout
import org.eclipse.ui.PartInitException
import org.eclipse.ui.PlatformUI
import org.eclipse.ui.editors.text.TextEditor
import org.eclipse.ui.part.FileEditorInput
import org.eclipse.ui.texteditor.SourceViewerDecorationSupport
import org.eclipse.ui.themes.ITheme
import org.eclipse.ui.themes.IThemeManager
import org.eclipse.ui.views.contentoutline.IContentOutlinePage
import de.isse.kiv.Settings
import de.isse.kiv.resources.Marker
import de.isse.kiv.resources.ResourceLookup
import de.isse.kiv.resources.ResourceProperties.toResourceProperties
import de.isse.kiv.source.Configuration
import de.isse.kiv.source.TokenMap
import de.isse.kiv.ui.Console
import de.isse.kiv.ui.outline.FileOutlinePage
import FileEditor._
import org.eclipse.ui.texteditor.ITextEditor
import de.isse.kiv.KIVPlugin
import org.eclipse.ui.IWorkbenchPart
import kiv.parser.Terminal
import org.eclipse.ui.actions.WorkspaceModifyOperation
import org.eclipse.core.resources.ResourcesPlugin
import org.eclipse.core.runtime.IProgressMonitor
import org.eclipse.core.resources.IWorkspaceRunnable
import org.eclipse.core.filesystem.IFileStore
import org.eclipse.ui.ide.IDE
import de.isse.kiv.KIVEditorServer
import java.nio.file.Paths
import de.isse.kiv.source.parser.KIVToken

object FileEditor {

  val FILE_EDITOR_ID = "de.isse.kiv.ui.editors.fileEditor"

  def openFile(file: IFile, cursor: Option[Range]) {
    if (!file.getProject.isOpen)
      file.getProject.open(null)
    val desc = PlatformUI.getWorkbench.getEditorRegistry.getDefaultEditor(file.getName)
    val page = ResourceLookup.activePage
    val editor = page.openEditor(new FileEditorInput(file), desc.getId).asInstanceOf[(ITextEditor)]
    cursor match {
      case Some(cursorRange) =>
        editor.selectAndReveal(cursorRange.start, cursorRange.length)
      case None =>
    }
  }
  def openFile(file: IFile) { openFile(file, None) }

  def openFile(filestore: IFileStore) {
    KIVEditorServer.openEditor(Paths.get(filestore.toURI()).toString)
  }

  def openFile(file: Either[IFile, IFileStore], cursor: Option[Range]) {
    file match {
      case Left(file) =>
        openFile(file, cursor)
      case Right(file) =>
        openFile(file) // TODO: currently ignores cursor
    }
  }
  def openFile(file: Either[IFile, IFileStore]) { openFile(file, None) }

  def selectRange(cursor: Option[Range]) {
    val cursorRange = cursor.getOrElse(return )
    ResourceLookup.activeTextEditor.selectAndReveal(cursorRange.start, cursorRange.length)
  }
}

class FileEditor extends TextEditorWithOutline() with IPropertyChangeListener {
  // NOTE: set the preference store, otherwise the bracket matcher does not work (i.e., is disabled,
  //       because the properties are not found in the wrong preference store)

  // setPreferenceStore(KIVPlugin.getDefault.getPreferenceStore)

  getPreferenceStore.setDefault(Settings.EDITOR_MATCHING_BRACKETS, true)
  getPreferenceStore.setDefault(Settings.EDITOR_MATCHING_BRACKETS_COLOR, Settings.EDITOR_MATCHING_BRACKETS_COLOR_VALUE)

  var markedToken: Option[KIVToken] = None

  val themeManager: IThemeManager = PlatformUI.getWorkbench().getThemeManager()
  val currentTheme: ITheme = themeManager.getCurrentTheme()
  val fontRegistry: FontRegistry = currentTheme.getFontRegistry()
  fontRegistry.addListener(this)

  val preferences = KIVPlugin.getDefault.getPreferenceStore
  preferences.addPropertyChangeListener(this)
  // var font: Font = fontRegistry.get("de.isse.kiv.ui.editors.editorTheme.fontDefinition")

  try {
    val page = ResourceLookup.activePage;
    page.showView(IPageLayout.ID_PROBLEM_VIEW)
  } catch {
    case e: PartInitException =>
  }

  Console.showConsole()
  setSourceViewerConfiguration(new Configuration(this))

  def getOutline = new FileOutlinePage(this);

  def getSelection = getSelectionProvider.getSelection match {
    case sel: ITextSelection => sel
    case _                   => null
  }

  def getFile =
    getEditorInput match {
      case input: FileEditorInput => input.getFile
      case _                      => null
    }

  def invalidateText() {
    object invalidateTextPresentation extends Runnable {
      def run() {
        val sourceViewer = getSourceViewer
        if (sourceViewer != null)
          sourceViewer.invalidateTextPresentation
      }
    }

    val display = Display.getDefault // getCurrent refers to current thread and may be null
    if (display != null)
      display.asyncExec(invalidateTextPresentation)
  }

  override protected def doSetInput(input: IEditorInput) {
    super.doSetInput(input)

    val file = getFile
    if (file == null) return

    val fileName = file.getName
    val index = fileName.lastIndexOf('.')
    val description = (if (index >= 0) fileName.substring(0, index) else fileName)
    val spec = file.getParent
    val specName = spec.getName
    setPartName(specName + " " + description)

    val model = file.fileModel
    if (model != null)
      model.observers += {
        _ => invalidateText
      }
  }

  protected override def configureSourceViewerDecorationSupport(support: SourceViewerDecorationSupport) {
    super.configureSourceViewerDecorationSupport(support)
    val matcher = new DefaultCharacterPairMatcher(Settings.EDITOR_BRACKET_MATCH_PAIRS)
    support.setCharacterPairMatcher(matcher)
    support.setMatchingCharacterPainterPreferenceKeys(Settings.EDITOR_MATCHING_BRACKETS, Settings.EDITOR_MATCHING_BRACKETS_COLOR)
  }

  override def propertyChange(event: PropertyChangeEvent) {
    setFont()
    TokenMap.readSyntaxhighlightFromPreferenceStore
    invalidateText

  }

  override def createPartControl(parent: Composite) {
    super.createPartControl(parent)
    setFont()

  }

  override def setBlockSelectionMode(enable: Boolean) {
    super.setBlockSelectionMode(enable)
    setFont()
  }

  def setFont() {
    val font = fontRegistry.get("de.isse.kiv.ui.editors.editorTheme.fontDefinition")
    val sourceviewer = getSourceViewer()
    if (sourceviewer != null)
      sourceviewer.getTextWidget().setFont(font)
  }

  override def handleCursorPositionChanged {
    super.handleCursorPositionChanged

    if (Settings.doHighlightVariables)
      markVariableOccurences()
  }

  def markVariableOccurences() {
    val selection = getSelection
    if (selection == null) return

    val file = getFile
    if (file == null) return

    val offset = getSelection.getOffset

    val token = getFile.fileModel.tokenAt(offset)
    if (token == markedToken) return

    markedToken = token
    Marker.deleteMarkers(getFile, Marker.TEXT_MARKER_ID)

    token match {
      case Some(token) if token.markOccurrences =>
        val tokens = getFile.fileModel.tokens

        val workspace = ResourcesPlugin.getWorkspace
        val rule = workspace.getRuleFactory.markerRule(file)

        object createMarkers extends WorkspaceModifyOperation(rule) with IWorkspaceRunnable {
          def execute(monitor: IProgressMonitor) {
            tokens.foreach { other =>
              if (other.markOccurrence(token))
                Marker.createVariableMarker(file, other.location.toRange)
            }
          }
        }

        workspace.run(createMarkers, null)
      case _ =>
    }
  }
}
