package de.isse.kiv.ui.handlers

import org.eclipse.core.commands.AbstractHandler
import org.eclipse.core.commands.ExecutionEvent
import de.isse.kiv.ui.Console
import org.eclipse.ui.PlatformUI
import org.eclipse.ui.texteditor.ITextEditor
import org.eclipse.jface.text.ITextSelection
import de.isse.kiv.resources.ResourceLookup
import org.eclipse.jface.text.IDocument
import CommentHandler._

object CommentHandler {
  val SINGLE_LINE_COMMAND = "de.isse.kiv.ui.commands.commentSingleLine"
  val MULTI_LINE_COMMAND = "de.isse.kiv.ui.commands.commentMultiLine"
}

class CommentHandler extends AbstractHandler {

  override def execute(event: ExecutionEvent): AnyRef = {
    try {
      PlatformUI.getWorkbench.getActiveWorkbenchWindow.getActivePage.getActiveEditor match {
        case editor: ITextEditor =>
          editor.getSelectionProvider.getSelection match {
            case selection: ITextSelection =>

              val doc = ResourceLookup.document
              var emptySelection = selection.getLength == 0
              //if nothing is selected, select line with cursor
              val (offset, length) =
                if (emptySelection)
                  getCursorSelectedLine(selection.getOffset, selection.getLength, doc)
                else
                  (selection.getOffset, selection.getLength)

              val command = event.getCommand.getId.toString

              command match {
                case SINGLE_LINE_COMMAND =>
                  handleSingleLineComments(offset, length, doc, editor, emptySelection)
                case MULTI_LINE_COMMAND =>
                  handleMultiLineComments(selection.getOffset, offset, length, doc, editor, emptySelection)
              }
          }
      }
    } catch {
      case e: Exception =>
    }

    null

  }

  def handleMultiLineComments(cursorOffset: Int, offset: Int, selectionLength: Int, doc: IDocument, editor: ITextEditor, isEmptySelection: Boolean) {
    val selectedString = doc.get(offset, selectionLength)

    val (commented, activeSelection, start, end, length) = isMultiLineCommented(cursorOffset, offset, selectionLength, doc)

    if (commented) {
      if (activeSelection) {
        val rpl = removeMultiLineCommentTokens(selectedString, start, end)
        replace(offset, selectionLength, rpl, editor, doc, isEmptySelection)

      } else {
        val pString = doc.get(start, length)
        val rpl = pString.drop(2).dropRight(2)
        replace(start, length, rpl, editor, doc, isEmptySelection)
      }

    } else {
      val rpl = addMultiLineCommentTokens(selectedString)
      replace(offset, selectionLength, rpl, editor, doc, isEmptySelection)
    }

  }

  def replace(offset: Int, selectionLength: Int, rpl: String, editor: ITextEditor, doc: IDocument, isEmptySelection: Boolean) {
    doc.replace(offset, selectionLength, rpl)
    if (!isEmptySelection)
      editor.selectAndReveal(offset, rpl.length)
    else if(rpl == "(::)"){
      //Sonderfall leerer Kommentarblock -> Cursor dazwischen setzen
      editor.selectAndReveal(offset+2, 0)
    }
  }

  def handleSingleLineComments(offset: Int, selectionLength: Int, doc: IDocument, editor: ITextEditor, isEmptySelection: Boolean) {
    val selectedString = doc.get(offset, selectionLength)
    val rpl = if (isSingleLineCommented(selectedString))
      removeSingleLineCommentTokens(selectedString)
    else
      addSingleLineCommentTokens(selectedString)
       
    replace(offset, selectionLength, rpl, editor, doc, isEmptySelection)

  }

  def getCursorSelectedLine(_offset: Int, _length: Int, doc: IDocument): (Int, Int) = {
    var offset = _offset
    var length = _length
    while (offset > 0 && doc.getChar(offset - 1) != '\n')
      offset -= 1
    length = 0
    while (offset + length < doc.getLength() && doc.getChar(offset + length) != '\n')
      length += 1
    //length -= 1
    (offset, length)

  }

  def removeSingleLineCommentTokens(s: String): String = {
    val splitted = s.linesWithSeparators
    val strb = new StringBuilder
    splitted.foreach { str =>
      if (str.startsWith(";; "))
        strb.append(str.drop(3))
      else if (str.startsWith(";;"))
        strb.append(str.drop(2))
      else
        strb.append(str)
    }
    strb.toString
  }

  def addSingleLineCommentTokens(s: String): String = {
    val splitted = s.linesWithSeparators
    val strb = new StringBuilder
    splitted.foreach { line =>
      strb.append(";; ")
      strb.append(line)
    }
    strb.toString
  }

  def isSingleLineCommented(s: String): Boolean = {
    s.split('\n').forall(str => str.trim.startsWith(";;") || str.trim.length == 0)

  }

  def addMultiLineCommentTokens(s: String): String = {
    return "(:" + s + ":)"
  }

  def removeMultiLineCommentTokens(s: String, start: Int, end: Int): String = {
    val (s1, s2) = s.splitAt(start)
    val (s3, s4) = s2.drop(2).splitAt(end - s1.length() - 2)
    return (s1 + s3 + s4.drop(2))
  }

  def isMultiLineCommented(cursorPosition: Int, offset: Int, selectionLength: Int, doc: IDocument): (Boolean, Boolean, Int, Int, Int) = {
    //Check if selection (if any) contains a (: :) Comment or else if the cursor is surrounded by comments
    if (selectionLength != 0) {
      //Selection
      val selectedString = doc.get(offset, selectionLength)
      //Did we select a commented area?

      val commentOpeningPosition = selectedString.indexOfSlice("(:")
      val commentClosingPosition = selectedString.lastIndexOfSlice(":)")
      val commented = (commentOpeningPosition < commentClosingPosition) && commentClosingPosition != -1 && commentOpeningPosition != -1
      if (commented) {
        //find position of surrounding comment tokens       
        return (true, true, commentOpeningPosition, commentClosingPosition, commentClosingPosition - commentOpeningPosition)
      }
    }
    //Cursor between? (: :)
    {
      val (s1, s2) = doc.get.splitAt(cursorPosition)
      val commentOpeningPosition = s1.lastIndexOfSlice("(:")
      val commentClosingPosition = s2.indexOfSlice(":)") + s1.length
      val commented = (commentOpeningPosition < commentClosingPosition) && commentClosingPosition != -1 && commentOpeningPosition != -1
      if (commented) {
        //find position of surrounding comment tokens       
        return (true, false, commentOpeningPosition, commentClosingPosition + 2, commentClosingPosition - commentOpeningPosition + 2)
      }
    }
    
    //Cursor left from? (: :)
    if(cursorPosition + 1 < doc.getLength)
    {
     
      val (s1, s2) = doc.get.splitAt(cursorPosition+2)
      val commentOpeningPosition = s1.lastIndexOfSlice("(:")
      val commentClosingPosition = s2.indexOfSlice(":)") + s1.length
      val commented = (commentOpeningPosition < commentClosingPosition) && commentClosingPosition != -1 && commentOpeningPosition != -1
      if (commented) {
        //find position of surrounding comment tokens    
        return (true, false, commentOpeningPosition, commentClosingPosition + 2, commentClosingPosition - commentOpeningPosition + 2)
      }
    }

    (false, false, -1, -1, -1)

  }

}
