package de.isse.kiv.ui.wizards

import java.io.File
import java.io.FileFilter
import java.io.IOException
import java.util.Arrays
import org.eclipse.core.internal.resources.ProjectDescriptionReader
import org.eclipse.core.resources.IProject
import org.eclipse.core.resources.IProjectDescription
import org.eclipse.core.resources.IWorkspaceRoot
import org.eclipse.core.resources.ResourcesPlugin
import org.eclipse.core.runtime.IPath
import org.eclipse.core.runtime.Path
import org.eclipse.jface.dialogs.Dialog
import org.eclipse.jface.wizard.WizardPage
import org.eclipse.swt.SWT
import org.eclipse.swt.events.ModifyEvent
import org.eclipse.swt.events.ModifyListener
import org.eclipse.swt.events.SelectionAdapter
import org.eclipse.swt.events.SelectionEvent
import org.eclipse.swt.layout.GridData
import org.eclipse.swt.layout.GridLayout
import org.eclipse.swt.widgets.Button
import org.eclipse.swt.widgets.Combo
import org.eclipse.swt.widgets.Composite
import org.eclipse.swt.widgets.DirectoryDialog
import org.eclipse.swt.widgets.Label
import org.eclipse.swt.widgets.Shell
import org.eclipse.swt.widgets.Table
import org.eclipse.swt.widgets.TableColumn
import org.eclipse.swt.widgets.TableItem
import org.eclipse.ui.ISharedImages
import org.eclipse.ui.PlatformUI
import de.isse.kiv.Settings
import de.isse.kiv.project.Nature
import de.isse.kiv.resources.ResourceLookup
import de.isse.kiv.resources.ResourceProperties
import de.isse.kiv.ui.Console
import ImportProjectWizardPage._
//remove if not needed
import scala.collection.JavaConversions._

object ImportProjectWizardPage {

  val SIZING_TEXT_FIELD_WIDTH = 250
}

class ImportProjectWizardPage(path:Option[String]) extends WizardPage("Import existing KIV Project") {

  var locationField: Combo = _

  var overwriteCheck: Button = _

  var overwriteLabel: Label = _

  var siblingProjectsTable: Table = _

  var siblingProjectsPath: IPath = _

  setTitle("Import existing KIV Project")

  setDescription("In order to import a KIV project select the root folder or any file within.")
  


  protected def validatePage(): Boolean = {
    val location = getProjectLocationPath
    if (locationField.isEnabled) {
      validateProjectPath(location)
    } else {
      setErrorMessage(null)
      true
    }
  }

  def validateProjectPath(location: IPath): Boolean = {
    val toplevel = location.toFile()
    if (!toplevel.exists()) {
      setErrorMessage("The specified project path does not exist.")
      overwriteCheck.setSelection(false)
      overwriteCheck.setEnabled(false)
      return false
    }
    if (!toplevel.isDirectory) {
      setErrorMessage("The specified project path is not a directory.")
      overwriteCheck.setSelection(false)
      overwriteCheck.setEnabled(false)
      return false
    }
    val devgraph_path = location.append(ResourceProperties.DEVGRAPH)
    val devgraph_file = devgraph_path.toFile()
    if (!devgraph_file.exists() || !devgraph_file.isFile) {
      setErrorMessage("The specified directory does not contain a KIV project (no devgraph found).")
      overwriteCheck.setSelection(false)
      overwriteCheck.setEnabled(false)
      return false
    }
    addProjectPath()
    val name = location.lastSegment()
    val root = ResourcesPlugin.getWorkspace.getRoot
    val project = root.getProject(name)
    if (project.exists()) {
      setErrorMessage("A project of the same name already exists.")
      overwriteCheck.setSelection(false)
      overwriteCheck.setEnabled(false)
      return false
    }
    val rootPath = root.getLocation
    val projectPath = rootPath.append(name)
    val projectFile = projectPath.append(IProjectDescription.DESCRIPTION_FILE_NAME)
    val pdr = new ProjectDescriptionReader()
    val isLocal = rootPath.isPrefixOf(location)
    if (!isLocal) {
      try {
        val pd = pdr.read(projectFile)
        if (pd.hasNature(Nature.NATURE_ID)) {
          var ok = true
          for (fn <- projectPath.toFile().list()) {
            ok &= (fn == IProjectDescription.DESCRIPTION_FILE_NAME || fn == ".settings")
          }
          if (!ok) {
            setErrorMessage("Cannot update links in an existing eclipse project.")
            overwriteCheck.setSelection(false)
            overwriteCheck.setEnabled(false)
            return false
          }
          val enabled = overwriteCheck.getEnabled
          if (!enabled) {
            overwriteCheck.setSelection(false)
            overwriteCheck.setEnabled(true)
          }
          val checked = overwriteCheck.getSelection
          if (!checked) {
            setErrorMessage("An eclipse project file already exists. Please confirm to update links.")
            return false
          }
        } else {
          overwriteCheck.setSelection(false)
          overwriteCheck.setEnabled(false)
        }
      } catch {
        case e: IOException =>
      }
    }
    setErrorMessage(null)
    true
  }

  private def getProjectLocationPath(): IPath = new Path(locationField.getText)

  def getProjects(): Array[IPath] = {
    val items = siblingProjectsTable.getSelection
    if (items == null || items.length == 0) {
      return Array(getProjectLocationPath)
    }
    val paths = Array.ofDim[IPath](items.length)
    for (i <- 0 until items.length) {
      paths(i) = siblingProjectsPath.append(items(i).getText)
    }
    paths
  }

  override def createControl(parent: Composite) {
    var data: GridData = null
    val columns = 4
    val composite = new Composite(parent, SWT.NULL)
    val layout = new GridLayout(columns, false)
    data = new GridData(GridData.FILL_HORIZONTAL)
    composite.setLayout(layout)
    composite.setLayoutData(data)
    val locationLabel = new Label(composite, SWT.NONE)
    locationLabel.setText("&Location")
    locationField = new Combo(composite, SWT.BORDER)
    locationField.setItems(Settings.projectLocations:_*)
    val i = locationField.indexOf(Settings.recentProjectLocation)
    locationField.select(if (i < 0) 0 else i)
    data = new GridData(GridData.FILL_HORIZONTAL)
    data.widthHint = SIZING_TEXT_FIELD_WIDTH
    data.horizontalSpan = 1
    locationField.setLayoutData(data)
    locationField.addModifyListener(new ModifyListener() {

      override def modifyText(e: ModifyEvent) {
        setPageComplete(validatePage())
      }
     locationField.setText(path.getOrElse(""))
    })
    val browseButton = new Button(composite, SWT.PUSH)
    browseButton.setText("B&rowse...")
    browseButton.addSelectionListener(new SelectionAdapter() {

      override def widgetSelected(event: SelectionEvent) {
        handleLocationBrowseButtonPressed()
      }
    })
    val removeLocationButton = new Button(composite, SWT.PUSH)
    removeLocationButton.setImage(ResourceLookup.stockImage(ISharedImages.IMG_TOOL_DELETE))
    removeLocationButton.setToolTipText("Delete current project location from history.")
    removeLocationButton.addSelectionListener(new SelectionAdapter() {

      override def widgetSelected(event: SelectionEvent) {
        removeCurrentLocation()
      }
    })
    overwriteCheck = new Button(composite, SWT.CHECK)
    overwriteCheck.setText("Confirm to update links in the existing Eclipse project")
    overwriteCheck.setEnabled(false)
    overwriteCheck.setSelection(false)
    overwriteCheck.addSelectionListener(new SelectionAdapter() {

      override def widgetSelected(event: SelectionEvent) {
        setPageComplete(validatePage())
      }
    })
    data = new GridData()
    data.horizontalSpan = columns
    overwriteCheck.setLayoutData(data)
    overwriteLabel = new Label(composite, SWT.NONE)
    overwriteLabel.setText("")
    data = new GridData()
    data.horizontalSpan = columns
    overwriteLabel.setLayoutData(data)
    data = new GridData(SWT.FILL, SWT.FILL, true, true)
    data.horizontalSpan = columns
    siblingProjectsTable = new Table(composite, SWT.BORDER | SWT.MULTI | SWT.FULL_SELECTION)
    siblingProjectsTable.setLayoutData(data)
    val column = new TableColumn(siblingProjectsTable, SWT.NONE)
    column.setText("Project")
    siblingProjectsTable.addSelectionListener(new SelectionAdapter() {

      override def widgetSelected(event: SelectionEvent) {
        val n = siblingProjectsTable.getSelectionCount
        if (n == 1) {
          val singleSelection = siblingProjectsTable.getSelection()(0).getText
          val path = siblingProjectsPath.append(singleSelection)
          locationField.setText(path.toString)
          locationField.setEnabled(true)
        }
        if (n > 1) {
          locationField.setText("Import " + n + " project" + (if (n != 1) "s" else "") + " from " + siblingProjectsPath)
          locationField.setEnabled(false)
        }
        setPageComplete(validatePage())
      }
    })
    setPageComplete(validatePage())
    setErrorMessage(null)
    setMessage(null)
    setControl(composite)
    Dialog.applyDialogFont(composite)
  }

  def addProjectPath() {
    val path = getProjectLocationPath
    if (path == null || path.isEmpty) return
    val parent = path.removeLastSegments(1)
    val dir = parent.toFile()
    if (dir == null || !dir.exists() || !dir.isDirectory) return
    if (parent != siblingProjectsPath) {
      siblingProjectsPath = parent
      refreshSiblingProjects(dir)
    }
    addCurrentLocation()
  }

  def refreshSiblingProjects(dir: File) {
    val filter = new FileFilter() {

      def accept(file: File): Boolean = {
        if (!file.isDirectory) return false
        var devgraphFile = new File(file.getAbsolutePath + File.separator + ResourceProperties.DEVGRAPH)
        return devgraphFile.exists()
      }
    }
    val files = dir.listFiles(filter).sorted
    if (files == null) return
    siblingProjectsTable.removeAll()
    for (file <- files) {
      val item = new TableItem(siblingProjectsTable, SWT.NONE)
      val name = file.getName
      item.setText(0, name)
    }
    val column = siblingProjectsTable.getColumn(0)
    column.pack()
    Console.debug.println("refresh path: " + dir.getAbsolutePath)
  }

  def addCurrentLocation() {
    val path = locationField.getText
    Settings.addProjectLocation(path)
    Settings.touchProjectLocation(path)
    var i: Int = 0
    i = locationField.indexOf(path)
    if (i < 0) locationField.add(path)
    Console.debug.println("add path: " + path)
  }

  def removeCurrentLocation() {
    val path = locationField.getText
    Settings.removeProjectLocation(path)
    locationField.remove(path)
    locationField.select(0)
    Console.debug.println("remove path: " + path)
  }

  def handleLocationBrowseButtonPressed() {
    val shell = PlatformUI.getWorkbench.getModalDialogShellProvider.getShell
    val dialog = new DirectoryDialog(shell)
    dialog.setFilterPath(locationField.getText)
    var selectedPath = dialog.open()
    if (selectedPath == null) selectedPath = ""
    locationField.setText(selectedPath)
  }
}
